# -*- mode: python; coding: utf-8 -*
# Copyright (c) 2019 Radio Astronomy Software Group
# Licensed under the 3-clause BSD License

import os
import fileinput

import pytest
import numpy as np
import warnings
from astropy import units
from astropy.coordinates import (
    SkyCoord,
    EarthLocation,
    Angle,
    AltAz,
    Longitude,
    Latitude,
)
from astropy.time import Time, TimeDelta
import scipy.io


from pyradiosky.data import DATA_PATH as SKY_DATA_PATH
from pyradiosky import utils as skyutils
from pyradiosky import skymodel

GLEAM_vot = os.path.join(SKY_DATA_PATH, "gleam_50srcs.vot")


@pytest.fixture
def time_location():
    array_location = EarthLocation(lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0)

    time = Time("2018-03-01 00:00:00", scale="utc", location=array_location)

    return time, array_location


@pytest.fixture
def zenith_skycoord(time_location):
    time, array_location = time_location

    source_coord = SkyCoord(
        alt=Angle(90, unit=units.deg),
        az=Angle(0, unit=units.deg),
        obstime=time,
        frame="altaz",
        location=array_location,
    )
    icrs_coord = source_coord.transform_to("icrs")

    return icrs_coord


@pytest.fixture
def zenith_skymodel(zenith_skycoord):
    icrs_coord = zenith_skycoord

    ra = icrs_coord.ra
    dec = icrs_coord.dec

    names = "zen_source"
    stokes = [1.0, 0, 0, 0]
    zenith_source = skymodel.SkyModel(names, ra, dec, stokes, "flat")

    return zenith_source


@pytest.fixture
def moonsky():
    pytest.importorskip("lunarsky")

    from lunarsky import MoonLocation, SkyCoord as SkyC

    # Tranquility base
    array_location = MoonLocation(lat="00d41m15s", lon="23d26m00s", height=0.0)

    time = Time.now()
    zen_coord = SkyC(
        alt=Angle(90, unit=units.deg),
        az=Angle(0, unit=units.deg),
        obstime=time,
        frame="lunartopo",
        location=array_location,
    )

    icrs_coord = zen_coord.transform_to("icrs")

    ra = icrs_coord.ra
    dec = icrs_coord.dec
    names = "zen_source"
    stokes = [1.0, 0, 0, 0]
    zenith_source = skymodel.SkyModel(names, ra, dec, stokes, "flat")

    zenith_source.update_positions(time, array_location)

    yield zenith_source


@pytest.fixture
def healpix_data():
    pytest.importorskip("astropy_healpix")
    import astropy_healpix

    nside = 32
    npix = astropy_healpix.nside_to_npix(nside)
    hp_obj = astropy_healpix.HEALPix(nside=nside)

    frequencies = np.linspace(100, 110, 10)
    pixel_area = astropy_healpix.nside_to_pixel_area(nside)

    # Note that the cone search includes any pixels that overlap with the search
    # region. With such a low resolution, this returns some slightly different
    # results from the equivalent healpy search. Subtracting(0.75 * pixres) from
    # the pixel area resolves this discrepancy for the test.

    pixres = hp_obj.pixel_resolution.to("deg").value
    ipix_disc = hp_obj.cone_search_lonlat(
        135 * units.deg, 0 * units.deg, radius=(10 - pixres * 0.75) * units.deg
    )

    return {
        "nside": nside,
        "npix": npix,
        "frequencies": frequencies,
        "pixel_area": pixel_area,
        "ipix_disc": ipix_disc,
    }


def test_source_zenith_from_icrs(time_location):
    """Test single source position at zenith constructed using icrs."""
    time, array_location = time_location

    lst = time.sidereal_time("apparent")

    tee_ra = lst
    cirs_ra = skyutils._tee_to_cirs_ra(tee_ra, time)

    cirs_source_coord = SkyCoord(
        ra=cirs_ra,
        dec=array_location.lat,
        obstime=time,
        frame="cirs",
        location=array_location,
    )

    icrs_coord = cirs_source_coord.transform_to("icrs")

    ra = icrs_coord.ra
    dec = icrs_coord.dec

    zenith_source = skymodel.SkyModel("icrs_zen", ra, dec, [1.0, 0, 0, 0], "flat")

    zenith_source.update_positions(time, array_location)
    zenith_source_lmn = zenith_source.pos_lmn.squeeze()
    assert np.allclose(zenith_source_lmn, np.array([0, 0, 1]), atol=1e-5)


def test_source_zenith(time_location, zenith_skymodel):
    """Test single source position at zenith."""
    time, array_location = time_location

    zenith_skymodel.update_positions(time, array_location)
    zenith_source_lmn = zenith_skymodel.pos_lmn.squeeze()
    assert np.allclose(zenith_source_lmn, np.array([0, 0, 1]))


def test_skymodel_init_errors(zenith_skycoord):
    icrs_coord = zenith_skycoord

    ra = icrs_coord.ra
    dec = icrs_coord.dec

    # Check error cases
    with pytest.raises(
        ValueError,
        match=(
            "UVParameter _ra is not the appropriate type. Is: float64. "
            "Should be: <class 'astropy.coordinates.angles.Longitude'>"
        ),
    ):
        skymodel.SkyModel("icrs_zen", ra.rad, dec, [1.0, 0, 0, 0], "flat")

    with pytest.raises(
        ValueError,
        match=(
            "UVParameter _dec is not the appropriate type. Is: float64. "
            "Should be: <class 'astropy.coordinates.angles.Latitude'>"
        ),
    ):
        skymodel.SkyModel("icrs_zen", ra, dec.rad, [1.0, 0, 0, 0], "flat")

    with pytest.raises(
        ValueError,
        match=(
            "Only one of freq_array and reference_frequency can be specified, not both."
        ),
    ):
        skymodel.SkyModel(
            "icrs_zen",
            ra,
            dec,
            [1.0, 0, 0, 0],
            "flat",
            reference_frequency=[1e8] * units.Hz,
            freq_array=[1e8] * units.Hz,
        )

    with pytest.raises(
        ValueError, match=("freq_array must have a unit that can be converted to Hz.")
    ):
        skymodel.SkyModel(
            "icrs_zen", ra, dec, [1.0, 0, 0, 0], "flat", freq_array=[1e8] * units.m
        )

    with pytest.raises(
        ValueError,
        match=("reference_frequency must have a unit that can be converted to Hz."),
    ):
        skymodel.SkyModel(
            "icrs_zen",
            ra,
            dec,
            [1.0, 0, 0, 0],
            "flat",
            reference_frequency=[1e8] * units.m,
        )


def test_skymodel_deprecated():
    """Test that old init works with deprecation."""
    source_new = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
        reference_frequency=np.array([1e8]) * units.Hz,
    )

    with pytest.warns(
        DeprecationWarning,
        match="The input parameters to SkyModel.__init__ have changed",
    ):
        source_old = skymodel.SkyModel(
            "Test",
            Longitude(12.0 * units.hr),
            Latitude(-30.0 * units.deg),
            [1.0, 0.0, 0.0, 0.0],
            np.array([1e8]) * units.Hz,
            "flat",
        )
    assert source_new == source_old

    with pytest.warns(
        DeprecationWarning,
        match="In the future, the reference_frequency will be required to be an astropy Quantity",
    ):
        source_old = skymodel.SkyModel(
            "Test",
            Longitude(12.0 * units.hr),
            Latitude(-30.0 * units.deg),
            [1.0, 0.0, 0.0, 0.0],
            "flat",
            reference_frequency=np.array([1e8]),
        )
    assert source_new == source_old

    source_old = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
        reference_frequency=np.array([1.5e8]) * units.Hz,
    )
    with pytest.warns(
        DeprecationWarning,
        match=(
            "Future equality does not pass, probably because the frequencies "
            "were not checked"
        ),
    ):
        assert source_new == source_old

    source_old = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg + 2e-3 * units.arcsec),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
        reference_frequency=np.array([1e8]) * units.Hz,
    )
    with pytest.warns(
        DeprecationWarning,
        match=("The _dec parameters are not within the future tolerance"),
    ):
        assert source_new == source_old

    source_old = skymodel.SkyModel(
        "Test",
        Longitude(Longitude(12.0 * units.hr) + Longitude(2e-3 * units.arcsec)),
        Latitude(-30.0 * units.deg),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
        reference_frequency=np.array([1e8]) * units.Hz,
    )
    with pytest.warns(
        DeprecationWarning,
        match=("The _ra parameters are not within the future tolerance"),
    ):
        assert source_new == source_old

    stokes = np.zeros((4, 2, 1), dtype=np.float)
    stokes[0, :, :] = 1.0
    source_new = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg),
        stokes,
        "subband",
        freq_array=np.array([1e8, 1.5e8]) * units.Hz,
    )
    with pytest.warns(
        DeprecationWarning,
        match="The input parameters to SkyModel.__init__ have changed",
    ):
        source_old = skymodel.SkyModel(
            "Test",
            Longitude(12.0 * units.hr),
            Latitude(-30.0 * units.deg),
            stokes,
            np.array([1e8, 1.5e8]) * units.Hz,
            "subband",
        )
    assert source_new == source_old

    with pytest.warns(
        DeprecationWarning,
        match="In the future, the freq_array will be required to be an astropy Quantity",
    ):
        source_old = skymodel.SkyModel(
            "Test",
            Longitude(12.0 * units.hr),
            Latitude(-30.0 * units.deg),
            stokes,
            "subband",
            freq_array=np.array([1e8, 1.5e8]),
        )
    assert source_new == source_old

    telescope_location = EarthLocation(
        lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0
    )
    time = Time("2018-01-01 00:00")
    with pytest.warns(
        DeprecationWarning,
        match="Passing telescope_location to SkyModel.coherency_calc is deprecated",
    ):
        source_new.update_positions(time, telescope_location)
        source_new.coherency_calc(telescope_location)


def test_update_position_errors(zenith_skymodel, time_location):
    time, array_location = time_location
    with pytest.raises(ValueError, match=("time must be an astropy Time object.")):
        zenith_skymodel.update_positions("2018-03-01 00:00:00", array_location)


def test_coherency_calc_errors():
    """Test that correct errors are raised when providing invalid location object."""
    coord = SkyCoord(ra=30.0 * units.deg, dec=40 * units.deg, frame="icrs")

    stokes_radec = [1, -0.2, 0.3, 0.1]

    source = skymodel.SkyModel("test", coord.ra, coord.dec, stokes_radec, "flat")

    with pytest.warns(UserWarning, match="Horizon cutoff undefined"):
        with pytest.raises(ValueError, match="telescope_location must be an"):
            source.coherency_calc().squeeze()


def test_calc_basis_rotation_matrix():
    """
    This tests whether the 3-D rotation matrix from RA/Dec to Alt/Az is
    actually a rotation matrix (R R^T = R^T R = I)
    """

    time = Time("2018-01-01 00:00")
    telescope_location = EarthLocation(
        lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0
    )

    source = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
    )
    source.update_positions(time, telescope_location)

    basis_rot_matrix = source._calc_average_rotation_matrix()

    assert np.allclose(np.matmul(basis_rot_matrix, basis_rot_matrix.T), np.eye(3))
    assert np.allclose(np.matmul(basis_rot_matrix.T, basis_rot_matrix), np.eye(3))


def test_calc_vector_rotation():
    """
    This checks that the 2-D coherency rotation matrix is unit determinant.
    I suppose we could also have checked (R R^T = R^T R = I)
    """

    time = Time("2018-01-01 00:00")
    telescope_location = EarthLocation(
        lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0
    )

    source = skymodel.SkyModel(
        "Test",
        Longitude(12.0 * units.hr),
        Latitude(-30.0 * units.deg),
        [1.0, 0.0, 0.0, 0.0],
        "flat",
    )
    source.update_positions(time, telescope_location)

    coherency_rotation = np.squeeze(source._calc_coherency_rotation())

    assert np.isclose(np.linalg.det(coherency_rotation), 1)


def test_pol_rotator():
    """
    Test that when above_horizon is unset, the coherency rotation is done for
    all polarized sources.
    """
    time = Time("2018-01-01 00:00")
    telescope_location = EarthLocation(
        lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0
    )

    Nsrcs = 50
    ras = Longitude(np.linspace(0, 24, Nsrcs) * units.hr)
    decs = Latitude(np.linspace(-90, 90, Nsrcs) * units.deg)
    names = np.arange(Nsrcs).astype("str")
    fluxes = np.array([[[5.5, 0.7, 0.3, 0.0]]] * Nsrcs).T

    # Make the last source non-polarized
    fluxes[..., -1] = [[1.0], [0], [0], [0]]

    source = skymodel.SkyModel(names, ras, decs, fluxes, "flat")

    assert source._n_polarized == Nsrcs - 1

    source.update_positions(time, telescope_location)

    # Check the default of inds for _calc_rotation_matrix()
    rots1 = source._calc_rotation_matrix()
    inds = np.array([25, 45, 16])
    rots2 = source._calc_rotation_matrix(inds)
    assert np.allclose(rots1[..., inds], rots2)

    # Unset the horizon mask and confirm that all rotation matrices are calculated.
    source.above_horizon = None

    with pytest.warns(UserWarning, match="Horizon cutoff undefined"):
        local_coherency = source.coherency_calc()

    # Check that all polarized sources are rotated.
    assert not np.all(
        np.isclose(local_coherency[..., :-1], source.coherency_radec[..., :-1])
    )
    assert np.allclose(local_coherency[..., -1], source.coherency_radec[..., -1])


def analytic_beam_jones(za, az, sigma=0.3):
    """
    Analytic beam with sensible polarization response.

    Required for testing polarized sources.
    """
    # B = np.exp(-np.tan(za/2.)**2. / 2. / sigma**2.)
    B = 1
    J = np.array(
        [[np.cos(za) * np.sin(az), np.cos(az)], [np.cos(az) * np.cos(za), -np.sin(az)]]
    )
    return B * J


def test_polarized_source_visibilities():
    """Test that visibilities of a polarized source match prior calculations."""
    array_location = EarthLocation(lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0)
    time0 = Time("2018-03-01 18:00:00", scale="utc", location=array_location)

    ha_off = 1 / 6.0
    ha_delta = 0.1
    time_offsets = np.arange(-ha_off, ha_off + ha_delta, ha_delta)
    zero_indx = np.argmin(np.abs(time_offsets))
    # make sure we get a true zenith time
    time_offsets[zero_indx] = 0.0
    times = time0 + time_offsets * units.hr
    ntimes = times.size

    zenith = SkyCoord(
        alt=90.0 * units.deg,
        az=0 * units.deg,
        frame="altaz",
        obstime=time0,
        location=array_location,
    )
    zenith_icrs = zenith.transform_to("icrs")

    src_astropy = SkyCoord(
        ra=zenith_icrs.ra, dec=zenith_icrs.dec, obstime=times, location=array_location
    )
    src_astropy_altaz = src_astropy.transform_to("altaz")
    assert np.isclose(src_astropy_altaz.alt.rad[zero_indx], np.pi / 2)

    stokes_radec = [1, -0.2, 0.3, 0.1]

    decoff = 0.0 * units.arcmin  # -0.17 * units.arcsec
    raoff = 0.0 * units.arcsec

    source = skymodel.SkyModel(
        "icrs_zen",
        Longitude(zenith_icrs.ra + raoff),
        Latitude(zenith_icrs.dec + decoff),
        stokes_radec,
        "flat",
    )

    coherency_matrix_local = np.zeros([2, 2, ntimes], dtype="complex128")
    alts = np.zeros(ntimes)
    azs = np.zeros(ntimes)
    for ti, time in enumerate(times):
        source.update_positions(time, telescope_location=array_location)
        alt, az = source.alt_az
        assert alt == src_astropy_altaz[ti].alt.radian
        assert az == src_astropy_altaz[ti].az.radian
        alts[ti] = alt
        azs[ti] = az

        coherency_tmp = source.coherency_calc().squeeze()
        coherency_matrix_local[:, :, ti] = coherency_tmp

    zas = np.pi / 2.0 - alts
    Jbeam = analytic_beam_jones(zas, azs)
    coherency_instr_local = np.einsum(
        "ab...,bc...,dc...->ad...", Jbeam, coherency_matrix_local, np.conj(Jbeam)
    )

    expected_instr_local = np.array(
        [
            [
                [
                    0.60632557 + 0.00000000e00j,
                    0.6031185 - 2.71050543e-20j,
                    0.60059597 + 0.00000000e00j,
                    0.59464231 + 5.42101086e-20j,
                    0.58939657 + 0.00000000e00j,
                ],
                [
                    0.14486082 + 4.99646382e-02j,
                    0.14776209 + 4.99943414e-02j,
                    0.14960097 + 5.00000000e-02j,
                    0.15302905 + 4.99773672e-02j,
                    0.15536376 + 4.99307015e-02j,
                ],
            ],
            [
                [
                    0.14486082 - 4.99646382e-02j,
                    0.14776209 - 4.99943414e-02j,
                    0.14960097 - 5.00000000e-02j,
                    0.15302905 - 4.99773672e-02j,
                    0.15536376 - 4.99307015e-02j,
                ],
                [
                    0.39282051 + 0.00000000e00j,
                    0.39674527 + 0.00000000e00j,
                    0.39940403 + 0.00000000e00j,
                    0.40481652 + 5.42101086e-20j,
                    0.40895287 + 0.00000000e00j,
                ],
            ],
        ]
    )

    assert np.allclose(coherency_instr_local, expected_instr_local)


def test_polarized_source_smooth_visibilities():
    """Test that visibilities change smoothly as a polarized source transits."""
    array_location = EarthLocation(lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0)
    time0 = Time("2018-03-01 18:00:00", scale="utc", location=array_location)

    ha_off = 1
    ha_delta = 0.01
    time_offsets = np.arange(-ha_off, ha_off + ha_delta, ha_delta)
    zero_indx = np.argmin(np.abs(time_offsets))
    # make sure we get a true zenith time
    time_offsets[zero_indx] = 0.0
    times = time0 + time_offsets * units.hr
    ntimes = times.size

    zenith = SkyCoord(
        alt=90.0 * units.deg,
        az=0 * units.deg,
        frame="altaz",
        obstime=time0,
        location=array_location,
    )
    zenith_icrs = zenith.transform_to("icrs")

    src_astropy = SkyCoord(
        ra=zenith_icrs.ra, dec=zenith_icrs.dec, obstime=times, location=array_location
    )
    src_astropy_altaz = src_astropy.transform_to("altaz")
    assert np.isclose(src_astropy_altaz.alt.rad[zero_indx], np.pi / 2)

    stokes_radec = [1, -0.2, 0.3, 0.1]

    source = skymodel.SkyModel(
        "icrs_zen", zenith_icrs.ra, zenith_icrs.dec, stokes_radec, "flat"
    )

    coherency_matrix_local = np.zeros([2, 2, ntimes], dtype="complex128")
    alts = np.zeros(ntimes)
    azs = np.zeros(ntimes)
    for ti, time in enumerate(times):
        source.update_positions(time, telescope_location=array_location)
        alt, az = source.alt_az
        assert alt == src_astropy_altaz[ti].alt.radian
        assert az == src_astropy_altaz[ti].az.radian
        alts[ti] = alt
        azs[ti] = az

        coherency_tmp = source.coherency_calc().squeeze()
        coherency_matrix_local[:, :, ti] = coherency_tmp

    zas = np.pi / 2.0 - alts
    Jbeam = analytic_beam_jones(zas, azs)
    coherency_instr_local = np.einsum(
        "ab...,bc...,dc...->ad...", Jbeam, coherency_matrix_local, np.conj(Jbeam)
    )

    # test that all the instrumental coherencies are smooth
    t_diff_sec = np.diff(times.jd) * 24 * 3600
    for pol_i in [0, 1]:
        for pol_j in [0, 1]:
            real_coherency = coherency_instr_local[pol_i, pol_j, :].real
            real_derivative = np.diff(real_coherency) / t_diff_sec
            real_derivative_diff = np.diff(real_derivative)
            assert np.max(np.abs(real_derivative_diff)) < 1e-6
            imag_coherency = coherency_instr_local[pol_i, pol_j, :].imag
            imag_derivative = np.diff(imag_coherency) / t_diff_sec
            imag_derivative_diff = np.diff(imag_derivative)
            assert np.max(np.abs(imag_derivative_diff)) < 1e-6

    # test that the stokes coherencies are smooth
    stokes_instr_local = skyutils.coherency_to_stokes(coherency_instr_local)
    for pol_i in range(4):
        real_stokes = stokes_instr_local[pol_i, :].real
        real_derivative = np.diff(real_stokes) / t_diff_sec
        real_derivative_diff = np.diff(real_derivative)
        assert np.max(np.abs(real_derivative_diff)) < 1e-6
        imag_stokes = stokes_instr_local[pol_i, :].imag
        assert np.all(imag_stokes == 0)


def test_read_healpix_hdf5(healpix_data):
    m = np.arange(healpix_data["npix"])
    m[healpix_data["ipix_disc"]] = healpix_data["npix"] - 1

    indices = np.arange(healpix_data["npix"])

    hpmap, inds, freqs = skymodel.read_healpix_hdf5(
        os.path.join(SKY_DATA_PATH, "healpix_disk.hdf5")
    )

    assert np.allclose(hpmap[0, :], m)
    assert np.allclose(inds, indices)
    assert np.allclose(freqs, healpix_data["frequencies"])


def test_healpix_to_sky(healpix_data):
    hpmap, inds, freqs = skymodel.read_healpix_hdf5(
        os.path.join(SKY_DATA_PATH, "healpix_disk.hdf5")
    )

    hmap_orig = np.arange(healpix_data["npix"])
    hmap_orig[healpix_data["ipix_disc"]] = healpix_data["npix"] - 1

    hmap_orig = np.repeat(hmap_orig[None, :], 10, axis=0)
    hmap_orig = (hmap_orig.T / skyutils.jy_to_ksr(freqs)).T
    hmap_orig = hmap_orig * healpix_data["pixel_area"]
    sky = skymodel.healpix_to_sky(hpmap, inds, freqs)

    assert np.allclose(sky.stokes[0], hmap_orig.value)


def test_units_healpix_to_sky(healpix_data):
    hpmap, inds, freqs = skymodel.read_healpix_hdf5(
        os.path.join(SKY_DATA_PATH, "healpix_disk.hdf5")
    )
    freqs = freqs * units.Hz

    brightness_temperature_conv = units.brightness_temperature(
        freqs, beam_area=healpix_data["pixel_area"]
    )
    stokes = (hpmap.T * units.K).to(units.Jy, brightness_temperature_conv).T
    sky = skymodel.healpix_to_sky(hpmap, inds, freqs)

    assert np.allclose(sky.stokes[0, 0], stokes.value[0])


def test_read_write_healpix(healpix_data):
    hpmap, inds, freqs = skymodel.read_healpix_hdf5(
        os.path.join(SKY_DATA_PATH, "healpix_disk.hdf5")
    )
    freqs = freqs * units.Hz
    filename = "tempfile.hdf5"
    with pytest.raises(ValueError) as verr:
        skymodel.write_healpix_hdf5(filename, hpmap, inds[:10], freqs.to("Hz").value)
    assert str(verr.value).startswith(
        "Need to provide nside if giving a subset of the map."
    )

    with pytest.raises(ValueError) as verr:
        skymodel.write_healpix_hdf5(
            filename,
            hpmap,
            inds[:10],
            freqs.to("Hz").value,
            nside=healpix_data["nside"],
        )
    assert str(verr.value).startswith("Invalid map shape")

    skymodel.write_healpix_hdf5(filename, hpmap, inds, freqs.to("Hz").value)

    hpmap_new, inds_new, freqs_new = skymodel.read_healpix_hdf5(filename)

    os.remove(filename)

    assert np.allclose(hpmap_new, hpmap)
    assert np.allclose(inds_new, inds)
    assert np.allclose(freqs_new, freqs.to("Hz").value)


def test_healpix_import_err():
    try:
        import astropy_healpix

        astropy_healpix.nside_to_npix(2 ** 3)
    except ImportError:
        errstr = "The astropy-healpix module must be installed to use HEALPix methods"
        Npix = 12
        hpmap = np.arange(Npix)
        inds = hpmap
        freqs = np.zeros(1)
        with pytest.raises(ImportError) as cm:
            skymodel.healpix_to_sky(hpmap, inds, freqs)
        assert str(cm.value).startswith(errstr)

        with pytest.raises(ImportError) as cm:
            skymodel.write_healpix_hdf5("filename.hdf5", hpmap, inds, freqs)
        assert str(cm.value).startswith(errstr)


def test_healpix_positions():
    pytest.importorskip("astropy_healpix")
    import astropy_healpix

    # write out a healpix file, read it back in check that it is as expected
    nside = 8
    Npix = astropy_healpix.nside_to_npix(nside)
    freqs = np.arange(100, 100.5, 0.1) * 1e6
    Nfreqs = len(freqs)
    hpx_map = np.zeros((Nfreqs, Npix))
    ipix = 357
    # Want 1 [Jy] converted to [K sr]
    hpx_map[:, ipix] = skyutils.jy_to_ksr(freqs)

    filename = os.path.join(SKY_DATA_PATH, "healpix_single.hdf5")

    skymodel.write_healpix_hdf5(filename, hpx_map, range(Npix), freqs)

    time = Time("2018-03-01 00:00:00", scale="utc")
    array_location = EarthLocation(lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0)

    ra, dec = astropy_healpix.healpix_to_lonlat(ipix, nside)
    skycoord_use = SkyCoord(ra, dec, frame="icrs")
    source_altaz = skycoord_use.transform_to(
        AltAz(obstime=time, location=array_location)
    )
    alt_az = np.array([source_altaz.alt.value, source_altaz.az.value])

    src_az = Angle(alt_az[1], unit="deg")
    src_alt = Angle(alt_az[0], unit="deg")
    src_za = Angle("90.d") - src_alt

    src_l = np.sin(src_az.rad) * np.sin(src_za.rad)
    src_m = np.cos(src_az.rad) * np.sin(src_za.rad)
    src_n = np.cos(src_za.rad)

    hpmap_hpx, indices_hpx, freqs_hpx = skymodel.read_healpix_hdf5(filename)
    os.remove(filename)

    sky_hpx = skymodel.healpix_to_sky(hpmap_hpx, indices_hpx, freqs_hpx)

    time.location = array_location

    sky_hpx.update_positions(time, array_location)
    src_alt_az = sky_hpx.alt_az
    assert np.isclose(src_alt_az[0][ipix], src_alt.rad)
    assert np.isclose(src_alt_az[1][ipix], src_az.rad)

    src_lmn = sky_hpx.pos_lmn
    assert np.isclose(src_lmn[0][ipix], src_l)
    assert np.isclose(src_lmn[1][ipix], src_m)
    assert np.isclose(src_lmn[2][ipix], src_n)


@pytest.mark.parametrize("spec_type", ["flat", "subband", "spectral_index", "full"])
def test_array_to_skymodel_loop(spec_type):
    if spec_type == "full":
        spectral_type = "subband"
    else:
        spectral_type = spec_type

    sky = skymodel.read_gleam_catalog(GLEAM_vot, spectral_type=spectral_type)
    if spec_type == "full":
        sky.spectral_type = "full"

    # This should be removed after pyuvdata PR #790 is merged.
    # GLEAM has NaNs for the spectral_index  of some sources
    # Currently, arrays with NaNs are never equal even if they are equal where
    # they are not nan and are nan in the same locations.
    # So remove those components for the round trip equality test.
    if spectral_type == "spectral_index":
        wh_not_nan = np.squeeze(np.argwhere(~np.isnan(sky.spectral_index)))
        sky.Ncomponents = wh_not_nan.size
        sky.ra = sky.ra[wh_not_nan]
        sky.dec = sky.dec[wh_not_nan]
        sky.name = sky.name[wh_not_nan]
        sky.reference_frequency = sky.reference_frequency[wh_not_nan]
        sky.spectral_index = sky.spectral_index[wh_not_nan]
        sky.stokes = sky.stokes[:, :, wh_not_nan]
        sky.coherency_radec = skyutils.stokes_to_coherency(sky.stokes)

    arr = skymodel.skymodel_to_array(sky)
    sky2 = skymodel.array_to_skymodel(arr)

    assert sky == sky2

    if spec_type == "flat":
        # again with no reference_frequency field
        reference_frequency = sky.reference_frequency
        sky.reference_frequency = None
        arr = skymodel.skymodel_to_array(sky)
        sky2 = skymodel.array_to_skymodel(arr)

        assert sky == sky2

        # again with flat & freq_array
        sky.freq_array = np.atleast_1d(np.unique(reference_frequency))
        arr = skymodel.skymodel_to_array(sky)
        sky2 = skymodel.array_to_skymodel(arr)

        assert sky == sky2


def test_param_flux_cuts():
    # Check that min/max flux limits in test params work.

    catalog_table = skymodel.read_votable_catalog(GLEAM_vot, return_table=True)

    catalog_table = skymodel.source_cuts(catalog_table, min_flux=0.2, max_flux=1.5)

    catalog = skymodel.array_to_skymodel(catalog_table)
    for sI in catalog.stokes[0, 0, :]:
        assert np.all(0.2 < sI < 1.5)


@pytest.mark.parametrize(
    "spec_type, init_kwargs, cut_kwargs",
    [
        ("flat", {}, {}),
        ("flat", {"reference_frequency": np.ones(20) * 200e6 * units.Hz}, {}),
        ("full", {"freq_array": np.array([1e8, 1.5e8]) * units.Hz}, {}),
        (
            "subband",
            {"freq_array": np.array([1e8, 1.5e8]) * units.Hz},
            {"freq_range": np.array([0.9e8, 2e8]) * units.Hz},
        ),
        (
            "subband",
            {"freq_array": np.array([1e8, 1.5e8]) * units.Hz},
            {"freq_range": np.array([1.1e8, 2e8]) * units.Hz},
        ),
        (
            "flat",
            {"freq_array": np.array([1e8]) * units.Hz},
            {"freq_range": np.array([0.9e8, 2e8]) * units.Hz},
        ),
    ],
)
def test_flux_cuts(spec_type, init_kwargs, cut_kwargs):
    Nsrcs = 20

    minflux = 0.5
    maxflux = 3.0

    ids = ["src{}".format(i) for i in range(Nsrcs)]
    ras = Longitude(np.random.uniform(0, 360.0, Nsrcs), units.deg)
    decs = Latitude(np.linspace(-90, 90, Nsrcs), units.deg)
    stokes = np.zeros((4, 1, Nsrcs), dtype=np.float)
    if spec_type == "flat":
        stokes[0, :, :] = np.linspace(minflux, maxflux, Nsrcs)
    else:
        stokes = np.zeros((4, 2, Nsrcs), dtype=np.float)
        stokes[0, 0, :] = np.linspace(minflux, maxflux / 2.0, Nsrcs)
        stokes[0, 1, :] = np.linspace(minflux * 2.0, maxflux, Nsrcs)

    skymodel_obj = skymodel.SkyModel(ids, ras, decs, stokes, spec_type, **init_kwargs)
    catalog_table = skymodel.skymodel_to_array(skymodel_obj)

    minI_cut = 1.0
    maxI_cut = 2.3

    cut_sourcelist = skymodel.source_cuts(
        catalog_table,
        latitude_deg=30.0,
        min_flux=minI_cut,
        max_flux=maxI_cut,
        **cut_kwargs,
    )

    if "freq_range" in cut_kwargs and np.min(
        cut_kwargs["freq_range"] > np.min(init_kwargs["freq_array"])
    ):
        assert np.all(cut_sourcelist["flux_density"][..., 0] < maxI_cut)
    else:
        assert np.all(cut_sourcelist["flux_density"][..., 0] > minI_cut)
        assert np.all(cut_sourcelist["flux_density"][..., 0] < maxI_cut)


@pytest.mark.parametrize(
    "spec_type, init_kwargs, cut_kwargs, error_category, error_message",
    [
        (
            "spectral_index",
            {
                "reference_frequency": np.ones(20) * 200e6 * units.Hz,
                "spectral_index": np.ones(20) * 0.8,
            },
            {},
            NotImplementedError,
            "Flux cuts with spectral index type objects is not supported yet.",
        ),
        (
            "full",
            {"freq_array": np.array([1e8, 1.5e8]) * units.Hz},
            {"freq_range": [0.9e8, 2e8]},
            ValueError,
            "freq_range must be an astropy Quantity.",
        ),
        (
            "subband",
            {"freq_array": np.array([1e8, 1.5e8]) * units.Hz},
            {"freq_range": 0.9e8 * units.Hz},
            ValueError,
            "freq_range must have 2 elements.",
        ),
        (
            "subband",
            {"freq_array": np.array([1e8, 1.5e8]) * units.Hz},
            {"freq_range": np.array([1.1e8, 1.4e8]) * units.Hz},
            ValueError,
            "No frequencies in freq_range.",
        ),
    ],
)
def test_source_cut_error(
    spec_type, init_kwargs, cut_kwargs, error_category, error_message
):
    Nsrcs = 20

    minflux = 0.5
    maxflux = 3.0

    ids = ["src{}".format(i) for i in range(Nsrcs)]
    ras = Longitude(np.random.uniform(0, 360.0, Nsrcs), units.deg)
    decs = Latitude(np.linspace(-90, 90, Nsrcs), units.deg)
    stokes = np.zeros((4, 1, Nsrcs), dtype=np.float)
    stokes = np.zeros((4, 1, Nsrcs), dtype=np.float)
    if spec_type == "flat" or spec_type == "spectral_index":
        stokes[0, :, :] = np.linspace(minflux, maxflux, Nsrcs)
    else:
        stokes = np.zeros((4, 2, Nsrcs), dtype=np.float)
        stokes[0, 0, :] = np.linspace(minflux, maxflux / 2.0, Nsrcs)
        stokes[0, 1, :] = np.linspace(minflux * 2.0, maxflux, Nsrcs)

    skymodel_obj = skymodel.SkyModel(ids, ras, decs, stokes, spec_type, **init_kwargs)
    catalog_table = skymodel.skymodel_to_array(skymodel_obj)

    minI_cut = 1.0
    maxI_cut = 2.3

    with pytest.raises(error_category, match=error_message):
        skymodel.source_cuts(
            catalog_table,
            latitude_deg=30.0,
            min_flux=minI_cut,
            max_flux=maxI_cut,
            **cut_kwargs,
        )


def test_circumpolar_nonrising(time_location):
    # Check that the source_cut function correctly identifies sources that are circumpolar or
    # won't rise.
    # Working with an observatory at the HERA latitude.

    time, location = time_location

    Ntimes = 100
    Nras = 20
    Ndecs = 20
    Nsrcs = Nras * Ndecs

    times = time + TimeDelta(np.linspace(0, 1.0, Ntimes), format="jd")
    lon = location.lon.deg
    ra = np.linspace(lon - 90, lon + 90, Nras)
    dec = np.linspace(-90, 90, Ndecs)

    ra, dec = map(np.ndarray.flatten, np.meshgrid(ra, dec))
    ra = Longitude(ra, units.deg)
    dec = Latitude(dec, units.deg)

    names = ["src{}".format(i) for i in range(Nsrcs)]
    stokes = np.zeros((4, 1, Nsrcs))
    stokes[0, ...] = 1.0

    sky = skymodel.SkyModel(names, ra, dec, stokes, "flat")

    src_arr = skymodel.skymodel_to_array(sky)
    src_arr = skymodel.source_cuts(src_arr, latitude_deg=location.lat.deg)

    # Boolean array identifying nonrising sources that were removed by source_cuts
    nonrising = np.array(
        [sky.name[ind] not in src_arr["source_id"] for ind in range(Nsrcs)]
    )
    is_below_horizon = np.zeros(Nsrcs).astype(bool)
    is_below_horizon[nonrising] = True

    alts, azs = [], []
    for ti in range(Ntimes):
        sky.update_positions(times[ti], location)
        alts.append(sky.alt_az[0])
        azs.append(sky.alt_az[1])

        # Check that sources below the horizon by coarse cut are
        # indeed below the horizon.
        lst = times[ti].sidereal_time("mean").rad
        dt0 = lst - src_arr["rise_lst"]
        dt1 = src_arr["set_lst"] - src_arr["rise_lst"]
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="invalid value encountered", category=RuntimeWarning
            )

            dt0[dt0 < 0] += 2 * np.pi
            dt1[dt1 < 0] += 2 * np.pi

            is_below_horizon[~nonrising] = dt0 > dt1
            assert np.all(sky.alt_az[0][is_below_horizon] < 0.0)

    alts = np.degrees(alts)

    # Check that the circumpolar and nonrising sources match expectation
    # from the tan(lat) * tan(dec) values.
    nonrising_test = np.where(np.all(alts < 0, axis=0))[0]
    circumpolar_test = np.where(np.all(alts > 0, axis=0))[0]

    tans = np.tan(location.lat.rad) * np.tan(dec.rad)
    nonrising_calc = np.where(tans < -1)
    circumpolar_calc = np.where(tans > 1)
    assert np.all(circumpolar_calc == circumpolar_test)
    assert np.all(nonrising_calc == nonrising_test)

    # Confirm that the source cuts excluded the non-rising sources.
    assert np.all(np.where(nonrising)[0] == nonrising_test)

    # check that rise_lst and set_lst get added to object when converted
    new_sky = skymodel.array_to_skymodel(src_arr)
    assert hasattr(new_sky, "_rise_lst")
    assert hasattr(new_sky, "_set_lst")

    # and that it's round tripped
    src_arr2 = skymodel.skymodel_to_array(new_sky)
    assert src_arr.dtype == src_arr2.dtype
    assert len(src_arr) == len(src_arr2)

    for name in src_arr.dtype.names:
        if isinstance(src_arr[name][0], (str,)):
            assert np.array_equal(src_arr[name], src_arr2[name])
        else:
            assert np.allclose(src_arr[name], src_arr2[name], equal_nan=True)


@pytest.mark.parametrize(
    "name_to_match, name_list, kwargs, result",
    [
        ("gle", ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"], {}, "GLEAM"),
        (
            "raj2000",
            ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"],
            {},
            "RAJ2000",
        ),
        (
            "ra",
            ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"],
            {"exclude_start_pattern": "_"},
            "RAJ2000",
        ),
        (
            "foo",
            ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"],
            {"brittle": False},
            None,
        ),
    ],
)
def test_get_matching_fields(name_to_match, name_list, kwargs, result):
    assert skymodel._get_matching_fields(name_to_match, name_list, **kwargs) == result


@pytest.mark.parametrize(
    "name_to_match, name_list, error_message",
    [
        (
            "j2000",
            ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"],
            "More than one match for j2000 in",
        ),
        (
            "foo",
            ["_RAJ2000", "_DEJ2000", "RAJ2000", "DEJ2000", "GLEAM"],
            "No match for foo in",
        ),
    ],
)
def test_get_matching_fields_errors(name_to_match, name_list, error_message):
    with pytest.raises(ValueError, match=error_message):
        skymodel._get_matching_fields(name_to_match, name_list)


@pytest.mark.parametrize("spec_type", ["flat", "subband"])
def test_read_gleam(spec_type):
    skymodel_obj = skymodel.read_gleam_catalog(GLEAM_vot, spectral_type=spec_type)

    assert skymodel_obj.Ncomponents == 50
    if spec_type == "subband":
        assert skymodel_obj.Nfreqs == 20

    # Check cuts
    source_select_kwds = {"min_flux": 0.5}
    cut_catalog = skymodel.read_gleam_catalog(
        GLEAM_vot,
        spectral_type=spec_type,
        source_select_kwds=source_select_kwds,
        return_table=True,
    )

    assert len(cut_catalog) < skymodel_obj.Ncomponents

    cut_obj = skymodel.read_gleam_catalog(
        GLEAM_vot, spectral_type=spec_type, source_select_kwds=source_select_kwds
    )

    assert len(cut_catalog) == cut_obj.Ncomponents

    source_select_kwds = {"min_flux": 10.0}
    with pytest.warns(UserWarning, match="All sources eliminated by cuts."):
        skymodel.read_gleam_catalog(
            GLEAM_vot,
            spectral_type=spec_type,
            source_select_kwds=source_select_kwds,
            return_table=True,
        )


def test_read_gleam_errors():
    with pytest.raises(ValueError, match="spectral_type full is not an allowed type"):
        skymodel.read_gleam_catalog(GLEAM_vot, spectral_type="full")


def test_read_deprecated_votable():
    votable_file = os.path.join(SKY_DATA_PATH, "single_source.vot")

    with pytest.warns(
        DeprecationWarning,
        match=(
            f"File {votable_file} contains tables with no name or ID, "
            "Support for such files is deprecated."
        ),
    ):
        skymodel_obj = skymodel.read_votable_catalog(votable_file)

    assert skymodel_obj.Ncomponents == 1

    with pytest.raises(ValueError, match=("More than one matching table.")):
        skymodel.read_votable_catalog(votable_file, id_column="de")


def test_read_votable_errors():

    # fmt: off
    flux_columns = ["Fint076", "Fint084", "Fint092", "Fint099", "Fint107",
                    "Fint115", "Fint122", "Fint130", "Fint143", "Fint151",
                    "Fint158", "Fint166", "Fint174", "Fint181", "Fint189",
                    "Fint197", "Fint204", "Fint212", "Fint220", "Fint227"]
    # fmt: on
    with pytest.raises(
        ValueError, match="freq_array must be provided for multiple flux columns."
    ):
        skymodel.read_votable_catalog(
            GLEAM_vot, flux_columns=flux_columns, reference_frequency=200e6 * units.Hz
        )

    with pytest.raises(
        ValueError, match="reference_frequency must be an astropy Quantity."
    ):
        skymodel.read_votable_catalog(GLEAM_vot, reference_frequency=200e6)


def test_idl_catalog_reader():
    catfile = os.path.join(SKY_DATA_PATH, "fhd_catalog.sav")
    sourcelist = skymodel.read_idl_catalog(catfile, expand_extended=False)

    catalog = scipy.io.readsav(catfile)["catalog"]
    assert len(sourcelist.ra) == len(catalog)


def test_idl_catalog_reader_extended_sources():
    catfile = os.path.join(SKY_DATA_PATH, "fhd_catalog.sav")
    sourcelist = skymodel.read_idl_catalog(catfile, expand_extended=True)

    catalog = scipy.io.readsav(catfile)["catalog"]
    ext_inds = np.where(
        [catalog["extend"][ind] is not None for ind in range(len(catalog))]
    )[0]
    ext_Ncomps = [len(catalog[ext]["extend"]) for ext in ext_inds]
    assert len(sourcelist.ra) == len(catalog) - len(ext_inds) + sum(ext_Ncomps)


def test_point_catalog_reader():
    catfile = os.path.join(SKY_DATA_PATH, "pointsource_catalog.txt")
    srcs = skymodel.read_text_catalog(catfile)

    with open(catfile, "r") as fhandle:
        header = fhandle.readline()
    header = [h.strip() for h in header.split()]
    dt = np.format_parser(
        ["U10", "f8", "f8", "f8", "f8"],
        ["source_id", "ra_j2000", "dec_j2000", "flux_density", "frequency"],
        header,
    )

    catalog_table = np.genfromtxt(
        catfile, autostrip=True, skip_header=1, dtype=dt.dtype
    )

    assert sorted(srcs.name) == sorted(catalog_table["source_id"])
    assert srcs.ra.deg in catalog_table["ra_j2000"]
    assert srcs.dec.deg in catalog_table["dec_j2000"]
    assert srcs.stokes in catalog_table["flux_density"]

    # Check cuts
    source_select_kwds = {"min_flux": 1.0}
    catalog = skymodel.read_text_catalog(
        catfile, source_select_kwds=source_select_kwds, return_table=True
    )
    assert len(catalog) == 2


def test_catalog_file_writer():
    time = Time(2458098.27471265, scale="utc", format="jd")
    array_location = EarthLocation(lat="-30d43m17.5s", lon="21d25m41.9s", height=1073.0)

    source_coord = SkyCoord(
        alt=Angle(90, unit=units.deg),
        az=Angle(0, unit=units.deg),
        obstime=time,
        frame="altaz",
        location=array_location,
    )
    icrs_coord = source_coord.transform_to("icrs")

    ra = icrs_coord.ra
    dec = icrs_coord.dec

    names = "zen_source"
    stokes = [1.0, 0, 0, 0]
    zenith_source = skymodel.SkyModel(names, ra, dec, stokes, "flat")

    fname = os.path.join(SKY_DATA_PATH, "temp_cat.txt")

    skymodel.write_catalog_to_file(fname, zenith_source)
    zenith_loop = skymodel.read_text_catalog(fname)
    assert np.all(zenith_loop == zenith_source)
    os.remove(fname)


@pytest.mark.parametrize("spec_type", ["flat", "subband", "spectral_index", "full"])
def test_text_catalog_loop(spec_type):
    if spec_type == "full":
        spectral_type = "subband"
    else:
        spectral_type = spec_type

    sky = skymodel.read_gleam_catalog(GLEAM_vot, spectral_type=spectral_type)
    if spec_type == "full":
        sky.spectral_type = "full"

    # This should be removed after pyuvdata PR #790 is merged.
    # GLEAM has NaNs for the spectral_index  of some sources
    # Currently, arrays with NaNs are never equal even if they are equal where
    # they are not nan and are nan in the same locations.
    # So remove those components for the round trip equality test.
    if spectral_type == "spectral_index":
        wh_not_nan = np.squeeze(np.argwhere(~np.isnan(sky.spectral_index)))
        sky.Ncomponents = wh_not_nan.size
        sky.ra = sky.ra[wh_not_nan]
        sky.dec = sky.dec[wh_not_nan]
        sky.name = sky.name[wh_not_nan]
        sky.reference_frequency = sky.reference_frequency[wh_not_nan]
        sky.spectral_index = sky.spectral_index[wh_not_nan]
        sky.stokes = sky.stokes[:, :, wh_not_nan]
        sky.coherency_radec = skyutils.stokes_to_coherency(sky.stokes)

    fname = os.path.join(SKY_DATA_PATH, "temp_cat.txt")
    skymodel.write_catalog_to_file(fname, sky)
    sky2 = skymodel.read_text_catalog(fname)
    sky_arr2 = skymodel.read_text_catalog(fname, return_table=True)
    sky3 = skymodel.array_to_skymodel(sky_arr2)
    os.remove(fname)

    assert sky == sky2
    assert sky == sky3

    if spec_type == "flat":
        # again with no reference_frequency field
        reference_frequency = sky.reference_frequency
        sky.reference_frequency = None
        arr = skymodel.skymodel_to_array(sky)
        sky2 = skymodel.array_to_skymodel(arr)

        assert sky == sky2

        # again with flat & freq_array
        sky.freq_array = np.atleast_1d(np.unique(reference_frequency))
        arr = skymodel.skymodel_to_array(sky)
        sky2 = skymodel.array_to_skymodel(arr)

        assert sky == sky2


@pytest.mark.parametrize("freq_mult", [1e-6, 1e-3, 1e3])
def test_text_catalog_loop_other_freqs(freq_mult):
    sky = skymodel.read_gleam_catalog(GLEAM_vot)
    sky.freq_array = np.atleast_1d(np.unique(sky.reference_frequency) * freq_mult)
    sky.reference_frequency = None

    fname = os.path.join(SKY_DATA_PATH, "temp_cat.txt")
    skymodel.write_catalog_to_file(fname, sky)
    sky2 = skymodel.read_text_catalog(fname)
    print(sky.spectral_type)
    print(sky2.spectral_type)
    os.remove(fname)

    assert sky == sky2


@pytest.mark.parametrize("spec_type", ["flat", "subband"])
def test_read_text_source_cuts(spec_type):

    sky = skymodel.read_gleam_catalog(GLEAM_vot, spectral_type=spec_type)
    fname = os.path.join(SKY_DATA_PATH, "temp_cat.txt")
    skymodel.write_catalog_to_file(fname, sky)

    source_select_kwds = {"min_flux": 0.5}
    cut_catalog = skymodel.read_text_catalog(
        fname, source_select_kwds=source_select_kwds, return_table=True,
    )

    assert len(cut_catalog) < sky.Ncomponents

    cut_obj = skymodel.read_text_catalog(fname, source_select_kwds=source_select_kwds)
    os.remove(fname)

    assert len(cut_catalog) == cut_obj.Ncomponents


def test_pyuvsim_mock_catalog_read():
    mock_cat_file = os.path.join(SKY_DATA_PATH, "mock_hera_text_2458098.27471.txt")

    mock_sky = skymodel.read_text_catalog(mock_cat_file)
    expected_names = ["src" + str(val) for val in np.arange(mock_sky.Ncomponents)]
    assert mock_sky.name.tolist() == expected_names


def test_read_text_errors():
    sky = skymodel.read_gleam_catalog(GLEAM_vot, spectral_type="subband")

    fname = os.path.join(SKY_DATA_PATH, "temp_cat.txt")
    skymodel.write_catalog_to_file(fname, sky)
    with fileinput.input(files=fname, inplace=True) as infile:
        for line in infile:
            line = line.replace("Flux_subband_76_MHz [Jy]", "Frequency [Hz]")
            print(line, end="")

    with pytest.raises(
        ValueError,
        match="If frequency column is present, only one flux column allowed.",
    ):
        skymodel.read_text_catalog(fname)

    skymodel.write_catalog_to_file(fname, sky)
    with fileinput.input(files=fname, inplace=True) as infile:
        for line in infile:
            line = line.replace("Flux_subband_76_MHz [Jy]", "Flux [Jy]")
            print(line, end="")

    with pytest.raises(
        ValueError,
        match="Multiple flux fields, but they do not all contain a frequency.",
    ):
        skymodel.read_text_catalog(fname)

    skymodel.write_catalog_to_file(fname, sky)
    with fileinput.input(files=fname, inplace=True) as infile:
        for line in infile:
            line = line.replace("SOURCE_ID", "NAME")
            print(line, end="")

    with pytest.raises(ValueError, match="Header does not match expectations."):
        skymodel.read_text_catalog(fname)

    os.remove(fname)


def test_zenith_on_moon(moonsky):
    """Source at zenith from the Moon."""

    zenith_source = moonsky
    zenith_source.check()

    zenith_source_lmn = zenith_source.pos_lmn.squeeze()
    assert np.allclose(zenith_source_lmn, np.array([0, 0, 1]))


def test_source_motion(moonsky):
    """ Check that period is about 28 days."""

    zenith_source = moonsky

    Ntimes = 500
    ets = np.linspace(0, 4 * 28 * 24 * 3600, Ntimes)
    times = zenith_source.time + TimeDelta(ets, format="sec")

    lmns = np.zeros((Ntimes, 3))
    for ti in range(Ntimes):
        zenith_source.update_positions(times[ti], zenith_source.telescope_location)
        lmns[ti] = zenith_source.pos_lmn.squeeze()
    _els = np.fft.fft(lmns[:, 0])
    dt = np.diff(ets)[0]
    _freqs = np.fft.fftfreq(Ntimes, d=dt)

    f_28d = 1 / (28 * 24 * 3600.0)

    maxf = _freqs[np.argmax(np.abs(_els[_freqs > 0]) ** 2)]
    assert np.isclose(maxf, f_28d, atol=2 / ets[-1])



# -*- mode: python; coding: utf-8 -*
# Copyright (c) 2019 Radio Astronomy Software Group
# Licensed under the 3-clause BSD License
"""Define SkyModel class and helper functions."""

import warnings

import numpy as np
from numpy.lib import recfunctions
from scipy.linalg import orthogonal_procrustes as ortho_procr
import scipy.io
from astropy.coordinates import Angle, EarthLocation, AltAz, Latitude, Longitude
from astropy.time import Time
import astropy.units as units
from astropy.units import Quantity
from astropy.io import votable
from pyuvdata.uvbase import UVBase
from pyuvdata.parameter import UVParameter

try:
    from pyuvdata.uvbeam.cst_beam import CSTBeam
except ImportError:  # pragma: no cover
    # backwards compatility for older pyuvdata versions
    from pyuvdata.cst_beam import CSTBeam

from . import utils as skyutils
from . import spherical_coords_transforms as sct


__all__ = [
    "hasmoon",
    "SkyModel",
    "read_healpix_hdf5",
    "healpix_to_sky",
    "skymodel_to_array",
    "array_to_skymodel",
    "source_cuts",
    "read_gleam_catalog",
    "read_votable_catalog",
    "read_text_catalog",
    "read_idl_catalog",
    "write_catalog_to_file",
    "write_healpix_hdf5",
]


try:
    from lunarsky import SkyCoord, MoonLocation, LunarTopo

    hasmoon = True
except ImportError:
    from astropy.coordinates import SkyCoord

    hasmoon = False

    class MoonLocation:
        pass

    class LunarTopo:
        pass


# Nov 5 2019 notes
#    Read/write methods to add:
#        FHD save file -- (read only)
#        VOTable -- (needs a write method)
#        HDF5 HEALPix --- (needs a write method)
#        HEALPix fits files

#    Convert stokes and coherency to Astropy quantities.


class SkyModel(UVBase):
    """
    Object to hold point source and diffuse models.

    Defines a set of components at given ICRS ra/dec coordinates,
    with flux densities defined by stokes parameters.

    Flux densities defined are by stokes parameters.
    The attribute Ncomponents gives the number of source components.

    Contains methods to:
        - Read and write different catalog formats.
        - Calculate source positions.
        - Calculate local coherency matrix in a local topocentric frame.

    Parameters
    ----------
    name : array_like of str
        Unique identifier for each source component, shape (Ncomponents,).
    ra : astropy Longitude object
        source RA in J2000 (or ICRS) coordinates, shape (Ncomponents,).
    dec : astropy Latitude object
        source Dec in J2000 (or ICRS) coordinates, shape (Ncomponents,).
    stokes : array_like of float
        4 element vector giving the source [I, Q, U, V], shape (4, Nfreqs, Ncomponents).
    spectral_type : str
        Indicates how fluxes should be calculated at each frequency.
        Options:
        - 'flat' : Flat spectrum.
        - 'full' : Flux is defined by a saved value at each frequency.
        - 'subband' : Flux is given at a set of band centers. (TODO)
        - 'spectral_index' : Flux is given at a reference frequency. (TODO)
    freq_array : astropy Quantity
        Array of frequencies that fluxes are provided for, shape (Nfreqs,).
    reference_frequency : astropy Quantity
        Reference frequencies of flux values, shape (Ncomponents,).
    spectral_index : array_like of float
        Spectral index of each source, shape (Ncomponents).
        None if spectral_type is not 'spectral_index'.
    pos_tol : float
        position tolerance in degrees, defaults to minimum float in numpy
        position tolerance in degrees
    extended_model_group : array_like of int
        Identifier that groups components of an extended source model.
        -1 for point sources, shape (Ncomponents,).
    beam_amp : array_like of float
        Beam amplitude at the source position, shape (4, Nfreqs, Ncomponents).
        4 element vector corresponds to [XX, YY, XY, YX] instrumental
        polarizations.

    """

    def __init__(
        self,
        name,
        ra,
        dec,
        stokes,
        spectral_type,
        freq_array=None,
        reference_frequency=None,
        spectral_index=None,
        pos_tol=np.finfo(float).eps,
        extended_model_group=None,
        beam_amp=None,
    ):
        # standard angle tolerance: 1 mas in radians.
        angle_tol = Angle(1, units.arcsec)
        self.future_angle_tol = Angle(1e-3, units.arcsec)

        self._Ncomponents = UVParameter(
            "Ncomponents", description="Number of components", expected_type=int
        )

        desc = (
            "Number of frequencies if spectral_type  is 'full' or 'subband', "
            "1 otherwise."
        )
        self._Nfreqs = UVParameter("Nfreqs", description=desc, expected_type=int)

        desc = "Right ascension of components in ICRS coordinates."
        self._ra = UVParameter(
            "ra",
            description=desc,
            form=("Ncomponents",),
            expected_type=Longitude,
            tols=angle_tol,
        )

        desc = "Declination of components in ICRS coordinates."
        self._dec = UVParameter(
            "dec",
            description=desc,
            form=("Ncomponents",),
            expected_type=Latitude,
            tols=angle_tol,
        )

        desc = "Component name."
        self._name = UVParameter(
            "name", description=desc, form=("Ncomponents",), expected_type=str
        )

        desc = "Frequency array in Hz, only required if spectral_type is 'full' or 'subband'."
        self._freq_array = UVParameter(
            "freq_array",
            description=desc,
            form=("Nfreqs",),
            expected_type=Quantity,
            required=False,
        )

        desc = "Reference frequency in Hz, only required if spectral_type is 'spectral_index'."
        self._reference_frequency = UVParameter(
            "reference_frequency",
            description=desc,
            form=("Ncomponents",),
            expected_type=Quantity,
            required=False,
        )

        desc = "Component flux per frequency and Stokes parameter"
        self._stokes = UVParameter(
            "stokes",
            description=desc,
            form=(4, "Nfreqs", "Ncomponents"),
            expected_type=float,
        )

        # The coherency is a 2x2 matrix giving electric field correlation in Jy
        self._coherency_radec = UVParameter(
            "coherency_radec",
            description="Ra/Dec coherency per component",
            form=(2, 2, "Nfreqs", "Ncomponents"),
            expected_type=np.complex,
        )

        desc = (
            "Type of spectral flux specification, options are: "
            "'full','flat', 'subband', 'spectral_index'."
        )
        self._spectral_type = UVParameter(
            "spectral_type",
            description=desc,
            expected_type=str,
            acceptable_vals=["full", "flat", "subband", "spectral_index"],
        )

        self._spectral_index = UVParameter(
            "spectral_index",
            description="Spectral indexm only required if spectral_type is 'spectral_index'.",
            form=("Ncomponents",),
            expected_type=float,
            required=False,
        )

        self._beam_amp = UVParameter(
            "beam_amp",
            description=(
                "Beam amplitude at the source position as a function "
                "of instrument polarization and frequency."
            ),
            form=(4, "Nfreqs", "Ncomponents"),
            expected_type=float,
            required=False,
        )

        self._extended_model_group = UVParameter(
            "extended_model_group",
            description=(
                "Identifier that groups components of an extended "
                "source model. Set to -1 for point sources."
            ),
            form=("Ncomponents",),
            expected_type=int,
            required=False,
        )

        desc = "Time for local position calculations."
        self._time = UVParameter(
            "time", description=desc, expected_type=Time, required=False
        )

        desc = "Telescope Location for local position calculations."
        self._telescope_location = UVParameter(
            "telescope_location",
            description=desc,
            expected_type=EarthLocation,
            required=False,
        )
        if hasmoon:
            self._telescope_location.expected_type = (EarthLocation, MoonLocation)

        desc = "Altitude and Azimuth of components in local coordinates."
        self._alt_az = UVParameter(
            "alt_az",
            description=desc,
            form=(2, "Ncomponents"),
            expected_type=float,
            tols=np.finfo(float).eps,
            required=False,
        )

        desc = "Position cosines of components in local coordinates."
        self._pos_lmn = UVParameter(
            "pos_lmn",
            description=desc,
            form=(3, "Ncomponents"),
            expected_type=float,
            tols=np.finfo(float).eps,
            required=False,
        )

        desc = (
            "Boolean indicator of whether this source is above the horizon"
            "at the current time."
            "True indicates the source is above the horizon."
        )
        self._above_horizon = UVParameter(
            "above_horizon",
            description=desc,
            form=("Ncomponents",),
            expected_type=bool,
            required=False,
        )

        # initialize the underlying UVBase properties
        super(SkyModel, self).__init__()

        self.name = np.atleast_1d(name)
        self.Ncomponents = self.name.size
        self.ra = np.atleast_1d(ra)
        self.dec = np.atleast_1d(dec)

        # handle old parameter order
        # (use to be: name, ra, dec, stokes, freq_array spectral_type)
        if isinstance(spectral_type, (np.ndarray, list, float, Quantity)):
            warnings.warn(
                "The input parameters to SkyModel.__init__ have changed. Please "
                "update the call.",
                category=DeprecationWarning,
            )
            freqs_use = spectral_type
            spectral_type = freq_array

            if spectral_type == "flat" and np.unique(freqs_use).size == 1:
                reference_frequency = np.zeros((self.Ncomponents), dtype=np.float)
                reference_frequency.fill(freqs_use[0])
                freq_array = None
            else:
                freq_array = freqs_use
                reference_frequency = None

        self.set_spectral_type_params(spectral_type)

        if freq_array is not None:
            if not isinstance(freq_array, (Quantity,)):
                warnings.warn(
                    "In the future, the freq_array will be required to be an "
                    "astropy Quantity with units that are convertable to Hz. "
                    "Currently, floats are assumed to be in Hz.",
                    category=DeprecationWarning,
                )
                freq_array = freq_array * units.Hz
            self.freq_array = np.atleast_1d(freq_array)
            self.Nfreqs = self.freq_array.size
        else:
            self.Nfreqs = 1

        if reference_frequency is not None:
            if not isinstance(reference_frequency, (Quantity,)):
                warnings.warn(
                    "In the future, the reference_frequency will be required to be an "
                    "astropy Quantity with units that are convertable to Hz. "
                    "Currently, floats are assumed to be in Hz.",
                    category=DeprecationWarning,
                )
                reference_frequency = reference_frequency * units.Hz
            self.reference_frequency = np.atleast_1d(reference_frequency)

        if spectral_index is not None:
            self.spectral_index = np.atleast_1d(spectral_index)

        self.stokes = np.asarray(stokes, dtype=np.float)
        if self.Ncomponents == 1:
            self.stokes = self.stokes.reshape(4, self.Nfreqs, 1)

        self._polarized = np.where(np.sum(self.stokes[1:, :, :], axis=0) != 0.0)[1]
        self._n_polarized = self._polarized.size

        self.coherency_radec = skyutils.stokes_to_coherency(self.stokes)

        self.check()

    def set_spectral_type_params(self, spectral_type):
        """Set parameters depending on spectral_type."""
        self.spectral_type = spectral_type

        if spectral_type == "spectral_index":
            self._spectral_index.required = True
            self._reference_frequency.required = True
            self._Nfreqs.acceptable_vals = [1]
            self._freq_array.required = False
        elif spectral_type in ["full", "subband"]:
            self._freq_array.required = True
            self._spectral_index.required = False
            self._reference_frequency.required = False
            self._Nfreqs.acceptable_vals = None
        elif spectral_type == "flat":
            self._freq_array.required = False
            self._spectral_index.required = False
            self._reference_frequency.required = False
            self._Nfreqs.acceptable_vals = [1]

    def check(self, check_extra=True, run_check_acceptability=True):
        """
        Check that all required parameters are set reasonably.

        Check that required parameters exist and have appropriate shapes.
        Optionally check if the values are acceptable.

        Parameters
        ----------
        check_extra : bool
            Option to check optional parameters as well as required ones.
        run_check_acceptability : bool
            Option to check if values in required parameters are acceptable.

        """
        # first make sure the required parameters and forms are set properly
        # for the spectral_type
        self.set_spectral_type_params(self.spectral_type)

        # make sure only one of freq_array and reference_frequency is defined
        if self.freq_array is not None and self.reference_frequency is not None:
            raise ValueError(
                "Only one of freq_array and reference_frequency can be specified, not both."
            )

        # Run the basic check from UVBase
        super(SkyModel, self).check(
            check_extra=check_extra, run_check_acceptability=run_check_acceptability
        )

        # make sure freq_array or reference_frequency if present is compatible with Hz
        if self.freq_array is not None:
            try:
                self.freq_array.to("Hz")
            except (units.UnitConversionError) as e:
                raise ValueError(
                    "freq_array must have a unit that can be converted to Hz."
                ) from e

        if self.reference_frequency is not None:
            try:
                self.reference_frequency.to("Hz")
            except (units.UnitConversionError) as e:
                raise ValueError(
                    "reference_frequency must have a unit that can be converted to Hz."
                ) from e

    def __eq__(self, other, check_extra=True):
        """Check for equality, check for future equality."""
        # Run the basic __eq__ from UVBase
        equal = super(SkyModel, self).__eq__(other, check_extra=check_extra)

        # Issue deprecation warning if ra/decs aren't close to future_angle_tol levels
        if not np.allclose(self.ra, other.ra, rtol=0, atol=self.future_angle_tol):
            warnings.warn(
                "The _ra parameters are not within the future tolerance. "
                f"Left is {self.ra}, right is {other.ra}",
                category=DeprecationWarning,
            )

        if not np.allclose(self.dec, other.dec, rtol=0, atol=self.future_angle_tol):
            warnings.warn(
                "The _dec parameters are not within the future tolerance. "
                f"Left is {self.dec}, right is {other.dec}",
                category=DeprecationWarning,
            )

        if not equal:
            warnings.warn(
                "Future equality does not pass, probably because the "
                "frequencies were not checked in the deprecated equality checking.",
                category=DeprecationWarning,
            )
            equal = super(SkyModel, self).__eq__(other, check_extra=False)

        return equal

    def _calc_average_rotation_matrix(self):
        """
        Calculate the "average" rotation matrix from RA/Dec to AltAz.

        This gets us close to the right value, then need to calculate a correction
        for each source separately.

        Returns
        -------
        array of floats
            Rotation matrix that defines the average mapping (RA,Dec) <--> (Alt,Az),
            shape (3, 3).
        """
        # unit vectors to be transformed by astropy
        x_c = np.array([1.0, 0, 0])
        y_c = np.array([0, 1.0, 0])
        z_c = np.array([0, 0, 1.0])

        axes_icrs = SkyCoord(
            x=x_c,
            y=y_c,
            z=z_c,
            obstime=self.time,
            location=self.telescope_location,
            frame="icrs",
            representation_type="cartesian",
        )

        axes_altaz = axes_icrs.transform_to("altaz")
        axes_altaz.representation_type = "cartesian"

        """ This transformation matrix is generally not orthogonal
            to better than 10^-7, so let's fix that. """

        R_screwy = axes_altaz.cartesian.xyz
        R_really_orthogonal, _ = ortho_procr(R_screwy, np.eye(3))

        # Note the transpose, to be consistent with calculation in sct
        R_really_orthogonal = np.array(R_really_orthogonal).T

        return R_really_orthogonal

    def _calc_rotation_matrix(self, inds=None):
        """
        Calculate the true rotation matrix from RA/Dec to AltAz for each component.

        Parameters
        ----------
        inds: array_like, optional
            Index array to select components.
            Defaults to all components.

        Returns
        -------
        array of floats
            Rotation matrix that defines the mapping (RA,Dec) <--> (Alt,Az),
            shape (3, 3, Ncomponents).
        """
        if inds is None:
            inds = range(self.Ncomponents)

        n_inds = len(inds)

        # Find mathematical points and vectors for RA/Dec
        theta_radec = np.pi / 2.0 - self.dec.rad[inds]
        phi_radec = self.ra.rad[inds]
        radec_vec = sct.r_hat(theta_radec, phi_radec)
        assert radec_vec.shape == (3, n_inds)

        # Find mathematical points and vectors for Alt/Az
        theta_altaz = np.pi / 2.0 - self.alt_az[0, inds]
        phi_altaz = self.alt_az[1, inds]
        altaz_vec = sct.r_hat(theta_altaz, phi_altaz)
        assert altaz_vec.shape == (3, n_inds)

        R_avg = self._calc_average_rotation_matrix()

        R_exact = np.zeros((3, 3, n_inds), dtype=np.float)

        for src_i in range(n_inds):
            intermediate_vec = np.matmul(R_avg, radec_vec[:, src_i])

            R_perturb = sct.vecs2rot(r1=intermediate_vec, r2=altaz_vec[:, src_i])

            R_exact[:, :, src_i] = np.matmul(R_perturb, R_avg)

        return R_exact

    def _calc_coherency_rotation(self, inds=None):
        """
        Calculate the rotation matrix to apply to the RA/Dec coherency to get it into alt/az.

        Parameters
        ----------
        inds: array_like, optional
            Index array to select components.
            Defaults to all components.

        Returns
        -------
        array of floats
            Rotation matrix that takes the coherency from (RA,Dec) --> (Alt,Az),
            shape (2, 2, Ncomponents).
        """
        if inds is None:
            inds = range(self.Ncomponents)
        n_inds = len(inds)

        basis_rotation_matrix = self._calc_rotation_matrix(inds)

        # Find mathematical points and vectors for RA/Dec
        theta_radec = np.pi / 2.0 - self.dec.rad[inds]
        phi_radec = self.ra.rad[inds]

        # Find mathematical points and vectors for Alt/Az
        theta_altaz = np.pi / 2.0 - self.alt_az[0, inds]
        phi_altaz = self.alt_az[1, inds]

        coherency_rot_matrix = np.zeros((2, 2, n_inds), dtype=np.float)
        for src_i in range(n_inds):
            coherency_rot_matrix[
                :, :, src_i
            ] = sct.spherical_basis_vector_rotation_matrix(
                theta_radec[src_i],
                phi_radec[src_i],
                basis_rotation_matrix[:, :, src_i],
                theta_altaz[src_i],
                phi_altaz[src_i],
            )

        return coherency_rot_matrix

    def coherency_calc(self, deprecated_location=None):
        """
        Calculate the local coherency in alt/az basis.

        SkyModel.update_positions() must be run prior to this method.

        The coherency is a 2x2 matrix giving electric field correlation in Jy.
        It's specified on the object as a coherency in the ra/dec basis,
        but must be rotated into local alt/az.

        Parameters
        ----------
        deprecated_location : EarthLocation object
            This keyword is deprecated. It is preserved to maintain backwards
            compatibility and sets the EarthLocation on this SkyModel object.

        Returns
        -------
        array of float
            local coherency in alt/az basis, shape (2, 2, Nfreqs, Ncomponents)
        """
        if self.above_horizon is None:
            warnings.warn(
                "Horizon cutoff undefined. Assuming all source components "
                "are above the horizon."
            )
            above_horizon = np.ones(self.Ncomponents).astype(bool)
        else:
            above_horizon = self.above_horizon

        if deprecated_location is not None:
            warnings.warn(
                "Passing telescope_location to SkyModel.coherency_calc is "
                "deprecated. Set the telescope_location via SkyModel.update_positions.",
                category=DeprecationWarning,
            )
            self.update_positions(self.time, deprecated_location)

        if not isinstance(self.telescope_location, (EarthLocation, MoonLocation)):

            errm = "telescope_location must be an astropy EarthLocation object"
            if hasmoon:
                errm += " or a lunarsky MoonLocation object "
            errm += ". "
            raise ValueError(
                errm + "value was: {al}".format(al=str(self.telescope_location))
            )

        # Select sources within the horizon only.
        coherency_local = self.coherency_radec[..., above_horizon]

        # For unpolarized sources, there's no need to rotate the coherency matrix.
        if self._n_polarized > 0:
            # If there are any polarized sources, do rotation.

            # This is a boolean array of length len(above_horizon)
            # that identifies polarized sources above the horizon.
            pol_over_hor = np.in1d(
                np.arange(self.Ncomponents)[above_horizon], self._polarized
            )

            # Indices of polarized sources in the full Ncomponents array,
            # downselected to those that are above the horizon.
            full_pol_over_hor = [pi for pi in self._polarized if above_horizon[pi]]

            if len(pol_over_hor) > 0:

                rotation_matrix = self._calc_coherency_rotation(full_pol_over_hor)

                rotation_matrix_T = np.swapaxes(rotation_matrix, 0, 1)
                coherency_local[:, :, :, pol_over_hor] = np.einsum(
                    "aby,bcxy,cdy->adxy",
                    rotation_matrix_T,
                    self.coherency_radec[:, :, :, full_pol_over_hor],
                    rotation_matrix,
                )

        return coherency_local

    def update_positions(self, time, telescope_location):
        """
        Calculate the altitude/azimuth positions for source components.

        From alt/az, calculate direction cosines (lmn)

        Doesn't return anything but updates the following attributes in-place:
        * ``pos_lmn``
        * ``alt_az``
        * ``time``

        Parameters
        ----------
        time : astropy Time object
            Time to update positions for.
        telescope_location : astropy EarthLocation object
            Telescope location to update positions for.
        """
        if not isinstance(time, Time):
            raise ValueError(
                "time must be an astropy Time object. value was: {t}".format(t=time)
            )

        if not isinstance(telescope_location, (EarthLocation, MoonLocation)):

            errm = "telescope_location must be an astropy EarthLocation object"
            if hasmoon:
                errm += " or a lunarsky MoonLocation object "
            errm += ". "
            raise ValueError(
                errm + "value was: {al}".format(al=str(telescope_location))
            )

        # Don't repeat calculations
        if self.time == time and self.telescope_location == telescope_location:
            return

        self.time = time
        self.telescope_location = telescope_location

        skycoord_use = SkyCoord(self.ra, self.dec, frame="icrs")
        if isinstance(self.telescope_location, MoonLocation):
            source_altaz = skycoord_use.transform_to(
                LunarTopo(obstime=self.time, location=self.telescope_location)
            )
        else:
            source_altaz = skycoord_use.transform_to(
                AltAz(obstime=self.time, location=self.telescope_location)
            )

        alt_az = np.array([source_altaz.alt.rad, source_altaz.az.rad])

        self.alt_az = alt_az

        pos_l = np.sin(alt_az[1, :]) * np.cos(alt_az[0, :])
        pos_m = np.cos(alt_az[1, :]) * np.cos(alt_az[0, :])
        pos_n = np.sin(alt_az[0, :])

        if self.pos_lmn is None:
            self.pos_lmn = np.zeros((3, self.Ncomponents), dtype=float)
        self.pos_lmn[0, :] = pos_l
        self.pos_lmn[1, :] = pos_m
        self.pos_lmn[2, :] = pos_n

        # Horizon mask:
        self.above_horizon = self.alt_az[0, :] > 0.0


def read_healpix_hdf5(hdf5_filename):
    """
    Read hdf5 healpix files using h5py and get a healpix map, indices and frequencies.

    Parameters
    ----------
    hdf5_filename : str
        Path and name of the hdf5 file to read.

    Returns
    -------
    hpmap : array_like of float
        Stokes-I surface brightness in K, for a set of pixels
        Shape (Ncomponents, Nfreqs)
    indices : array_like, int
        Corresponding HEALPix indices for hpmap.
    freqs : array_like, float
        Frequencies in Hz. Shape (Nfreqs)
    """
    import h5py

    with h5py.File(hdf5_filename, "r") as file:
        hpmap = file["data"][0, ...]  # Remove Nskies axis.
        indices = file["indices"][()]
        freqs = file["freqs"][()]
    return hpmap, indices, freqs


def write_healpix_hdf5(filename, hpmap, indices, freqs, nside=None, history=None):
    """
    Write a set of HEALPix maps to an HDF5 file.

    Parameters
    ----------
    filename: str
        Name of file to write to.
    hpmap: array_like of float
        Pixel values in Kelvin. Shape (Nfreqs, Npix)
    indices: array_like of int
        HEALPix pixel indices corresponding with axis 1 of hpmap.
    freqs: array_like of floats
        Frequencies in Hz corresponding with axis 0 of hpmap.
    nside: int
        nside parameter of the map. Optional if the hpmap covers
        the full sphere (i.e., has no missing pixels), since the nside
        can be inferred from the map size.
    history: str
        Optional history string to include in the file.

    """
    try:
        import astropy_healpix
    except ImportError as e:
        raise ImportError(
            "The astropy-healpix module must be installed to use HEALPix methods"
        ) from e

    import h5py

    Nfreqs = freqs.size
    Npix = len(indices)
    if nside is None:
        try:
            nside = astropy_healpix.npix_to_nside(Npix)
        except ValueError:
            raise ValueError("Need to provide nside if giving a subset of the map.")

    try:
        assert hpmap.shape == (Nfreqs, Npix)
    except AssertionError:
        raise ValueError("Invalid map shape {}".format(str(hpmap.shape)))

    if history is None:
        history = ""

    valid_params = {
        "Npix": Npix,
        "nside": nside,
        "Nskies": 1,
        "Nfreqs": Nfreqs,
        "data": hpmap[None, ...],
        "indices": indices,
        "freqs": freqs,
        "history": history,
    }
    dsets = {
        "data": np.float64,
        "indices": np.int32,
        "freqs": np.float64,
        "history": h5py.special_dtype(vlen=str),
    }

    with h5py.File(filename, "w") as fileobj:
        for k in valid_params:
            d = valid_params[k]
            if k in dsets:
                if np.isscalar(d):
                    fileobj.create_dataset(k, data=d, dtype=dsets[k])
                else:
                    fileobj.create_dataset(
                        k,
                        data=d,
                        dtype=dsets[k],
                        compression="gzip",
                        compression_opts=9,
                    )
            else:
                fileobj.attrs[k] = d


def healpix_to_sky(hpmap, indices, freqs):
    """
    Convert a healpix map in K to a set of point source components in Jy.

    Parameters
    ----------
    hpmap : array_like of float
        Stokes-I surface brightness in K, for a set of pixels
        Shape (Nfreqs, Ncomponents)
    indices : array_like, int
        Corresponding HEALPix indices for hpmap.
    freqs : array_like, float
        Frequencies in Hz. Shape (Nfreqs)

    Returns
    -------
    sky : :class:`SkyModel`
        The sky model created from the healpix map.

    Notes
    -----
    Currently, this function only converts a HEALPix map with a frequency axis.
    """
    try:
        import astropy_healpix
    except ImportError as e:
        raise ImportError(
            "The astropy-healpix module must be installed to use HEALPix methods"
        ) from e

    nside = astropy_healpix.npix_to_nside(hpmap.shape[-1])
    ra, dec = astropy_healpix.healpix_to_lonlat(indices, nside)
    freq = Quantity(freqs, "hertz")
    stokes = np.zeros((4, len(freq), len(indices)))
    stokes[0] = (hpmap.T / skyutils.jy_to_ksr(freq)).T
    stokes[0] = stokes[0] * astropy_healpix.nside_to_pixel_area(nside)

    sky = SkyModel(indices.astype("str"), ra, dec, stokes, "full", freq_array=freq)
    return sky


def skymodel_to_array(sky):
    """
    Make a recarray of source components from a SkyModel object.

    Parameters
    ----------
    sky : :class:`pyradiosky.SkyModel`
        SkyModel object to convert to a recarray.

    Returns
    -------
    catalog_table : recarray
        recarray equivalent to SkyModel data.

    """
    sky.check()
    max_name_len = np.max([len(name) for name in sky.name])
    fieldtypes = ["U" + str(max_name_len), "f8", "f8", "f8"]
    fieldnames = ["source_id", "ra_j2000", "dec_j2000", "flux_density"]
    fieldshapes = [()] * 3

    n_stokes = 4

    if sky.freq_array is not None:
        if sky.spectral_type == "subband":
            fieldnames.append("subband_frequency")
        else:
            fieldnames.append("frequency")
        fieldtypes.append("f8")
        fieldshapes.extend([(sky.Nfreqs, n_stokes,), (sky.Nfreqs,)])
    elif sky.reference_frequency is not None:
        # add frequency field (a copy of reference_frequency) for backwards
        # compatibility.
        warnings.warn(
            "The frequency field is included in the recarray as a copy of the "
            "reference_frequency field for backwards compatibility. In future "
            "only the reference_frequency will be included.",
            category=DeprecationWarning,
        )
        fieldnames.extend(["frequency", "reference_frequency"])
        fieldtypes.extend(["f8"] * 2)
        fieldshapes.extend([(n_stokes,)] + [()] * 2)
        if sky.spectral_index is not None:
            fieldnames.append("spectral_index")
            fieldtypes.append("f8")
            fieldshapes.append(())
    else:
        # flat spectrum, no freq info
        fieldshapes.append((n_stokes,))

    if hasattr(sky, "_rise_lst"):
        fieldnames.append("rise_lst")
        fieldtypes.append("f8")
        fieldshapes.append(())

    if hasattr(sky, "_set_lst"):
        fieldnames.append("set_lst")
        fieldtypes.append("f8")
        fieldshapes.append(())

    dt = np.dtype(list(zip(fieldnames, fieldtypes, fieldshapes)))

    arr = np.empty(sky.Ncomponents, dtype=dt)
    arr["source_id"] = sky.name
    arr["ra_j2000"] = sky.ra.deg
    arr["dec_j2000"] = sky.dec.deg
    if sky.freq_array is not None:
        if sky.spectral_type == "subband":
            arr["subband_frequency"] = sky.freq_array
        else:
            arr["frequency"] = sky.freq_array
        arr["flux_density"] = sky.stokes.T
    elif sky.reference_frequency is not None:
        arr["reference_frequency"] = sky.reference_frequency
        arr["frequency"] = sky.reference_frequency
        arr["flux_density"] = np.squeeze(sky.stokes.T)
        if sky.spectral_index is not None:
            arr["spectral_index"] = sky.spectral_index
    else:
        # flat spectral type, no freq info
        arr["flux_density"] = np.squeeze(sky.stokes.T)

    if hasattr(sky, "_rise_lst"):
        arr["rise_lst"] = sky._rise_lst
    if hasattr(sky, "_set_lst"):
        arr["set_lst"] = sky._set_lst

    return arr


def array_to_skymodel(catalog_table):
    """
    Make a SkyModel object from a recarray.

    Parameters
    ----------
    catalog_table : recarray
        recarray to turn into a SkyModel object.

    Returns
    -------
    :class:`pyradiosky.SkyModel`

    """
    ra = Longitude(catalog_table["ra_j2000"], units.deg)
    dec = Latitude(catalog_table["dec_j2000"], units.deg)
    ids = np.asarray(catalog_table["source_id"]).astype(str)
    stokes = np.atleast_1d(catalog_table["flux_density"]).T

    rise_lst = None
    set_lst = None

    fieldnames = catalog_table.dtype.names
    if "reference_frequency" in fieldnames:
        reference_frequency = Quantity(
            np.atleast_1d(catalog_table["reference_frequency"]), "hertz"
        )
        if "spectral_index" in fieldnames:
            spectral_index = np.atleast_1d(catalog_table["spectral_index"])
            spectral_type = "spectral_index"
        else:
            spectral_type = "flat"
            spectral_index = None
        freq_array = None
    elif "frequency" in fieldnames or "subband_frequency" in fieldnames:
        if "frequency" in fieldnames:
            freq_array = Quantity(np.atleast_1d(catalog_table["frequency"]), "hertz")
        else:
            spectral_type = "subband"
            freq_array = Quantity(
                np.atleast_1d(catalog_table["subband_frequency"]), "hertz"
            )
        # freq_array gets copied for every component, so its zeroth axis is
        # length Ncomponents. Just take the first one.
        freq_array = freq_array[0, :]
        if freq_array.size > 1:
            if "subband_frequency" not in fieldnames:
                spectral_type = "full"
        else:
            spectral_type = "flat"
        reference_frequency = None
        spectral_index = None
    else:
        # flat spectrum, no freq info
        spectral_type = "flat"
        freq_array = None
        reference_frequency = None
        spectral_index = None

    if "rise_lst" in catalog_table.dtype.names:
        rise_lst = catalog_table["rise_lst"]
        set_lst = catalog_table["set_lst"]

    if stokes.ndim == 2:
        stokes = stokes[:, np.newaxis, :]  # Add frequency axis if there isn't one.

    skymodel = SkyModel(
        ids,
        ra,
        dec,
        stokes,
        spectral_type,
        freq_array=freq_array,
        reference_frequency=reference_frequency,
        spectral_index=spectral_index,
    )

    if rise_lst is not None:
        skymodel._rise_lst = rise_lst
    if set_lst is not None:
        skymodel._set_lst = set_lst

    return skymodel


def source_cuts(
    catalog_table,
    latitude_deg=None,
    horizon_buffer=0.04364,
    min_flux=None,
    max_flux=None,
    freq_range=None,
):
    """
    Perform flux and horizon selections on recarray of source components.

    Parameters
    ----------
    catalog_table : recarray
        recarray of source catalog information. Must have the columns:
        'source_id', 'ra_j2000', 'dec_j2000', 'flux_density'
        may also have the colums:
        'frequency' or 'reference_frequency'
    latitude_deg : float
        Latitude of telescope in degrees. Used to estimate rise/set lst.
    horizon_buffer : float
        Angle buffer for coarse horizon cut in radians.
        Default is about 10 minutes of sky rotation. `SkyModel`
        components whose calculated altitude is less than `horizon_buffer` are excluded.
        Caution! The altitude calculation does not account for precession/nutation of the Earth.
        The buffer angle is needed to ensure that the horizon cut doesn't exclude sources near
        but above the horizon. Since the cutoff is done using lst, and the lsts are calculated
        with astropy, the required buffer should _not_ drift with time since the J2000 epoch.
        The default buffer has been tested around julian date 2457458.0.
    min_flux : float
        Minimum stokes I flux to select [Jy]
    max_flux : float
        Maximum stokes I flux to select [Jy]
    freq_range : astropy Quantity
        Frequency range over which the min and max flux tests should be performed.
        Must be length 2. If None, use the range over which the object is defined.

    Returns
    -------
    recarray
        A new recarray of source components, with additional columns for rise and set lst.
    """
    coarse_horizon_cut = latitude_deg is not None

    if coarse_horizon_cut:
        lat_rad = np.radians(latitude_deg)
        buff = horizon_buffer

    if freq_range is not None:
        if not isinstance(freq_range, (Quantity,)):
            raise ValueError("freq_range must be an astropy Quantity.")
        if not np.atleast_1d(freq_range).size == 2:
            raise ValueError("freq_range must have 2 elements.")

    fieldnames = catalog_table.dtype.names
    if min_flux or max_flux:
        if "reference_frequency" in fieldnames:
            if "spectral_index" in fieldnames:
                raise NotImplementedError(
                    "Flux cuts with spectral index type objects is not supported yet."
                )
            else:
                # flat spectral type
                if min_flux:
                    catalog_table = catalog_table[
                        catalog_table["flux_density"][..., 0] > min_flux
                    ]
                if max_flux:
                    catalog_table = catalog_table[
                        catalog_table["flux_density"][..., 0] < max_flux
                    ]
        elif "frequency" in fieldnames or "subband_frequency" in fieldnames:
            if "frequency" in fieldnames:
                freq_array = Quantity(
                    np.atleast_1d(catalog_table["frequency"]), "hertz"
                )
            else:
                freq_array = Quantity(
                    np.atleast_1d(catalog_table["subband_frequency"]), "hertz"
                )
            # freq_array gets copied for every component, so its zeroth axis is
            # length Ncomponents. Just take the first one.
            freq_array = freq_array[0, :]
            if freq_range is not None:
                freqs_inds_use = np.where(
                    (freq_array >= np.min(freq_range))
                    & (freq_array <= np.max(freq_range))
                )[0]
                if freqs_inds_use.size == 0:
                    raise ValueError("No frequencies in freq_range.")
            else:
                freqs_inds_use = np.arange(freq_array.size)
            flux_data = np.atleast_1d(catalog_table["flux_density"][..., 0])
            if flux_data.ndim > 1:
                flux_data = flux_data[:, freqs_inds_use]
            if min_flux:
                row_inds = np.where(np.min(flux_data, axis=1) > min_flux)
                catalog_table = catalog_table[row_inds]
                flux_data = flux_data[row_inds]
            if max_flux:
                catalog_table = catalog_table[
                    np.where(np.max(flux_data, axis=1) < max_flux)
                ]
        else:
            # flat spectral type
            if min_flux:
                catalog_table = catalog_table[
                    catalog_table["flux_density"][..., 0] > min_flux
                ]
            if max_flux:
                catalog_table = catalog_table[
                    catalog_table["flux_density"][..., 0] < max_flux
                ]

    ra = Longitude(catalog_table["ra_j2000"], units.deg)
    dec = Latitude(catalog_table["dec_j2000"], units.deg)

    if coarse_horizon_cut:
        tans = np.tan(lat_rad) * np.tan(dec.rad)
        nonrising = tans < -1

        catalog_table = catalog_table[~nonrising]
        tans = tans[~nonrising]

        ra = ra[~nonrising]

    if coarse_horizon_cut:
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="invalid value encountered", category=RuntimeWarning
            )
            rise_lst = ra.rad - np.arccos((-1) * tans) - buff
            set_lst = ra.rad + np.arccos((-1) * tans) + buff

            rise_lst[rise_lst < 0] += 2 * np.pi
            set_lst[set_lst < 0] += 2 * np.pi
            rise_lst[rise_lst > 2 * np.pi] -= 2 * np.pi
            set_lst[set_lst > 2 * np.pi] -= 2 * np.pi

        catalog_table = recfunctions.append_fields(
            catalog_table, ["rise_lst", "set_lst"], [rise_lst, set_lst], usemask=False
        )

    if len(catalog_table) == 0:
        warnings.warn("All sources eliminated by cuts.")

    return catalog_table


def _get_matching_fields(
    name_to_match, name_list, exclude_start_pattern=None, brittle=True
):
    match_list = [name for name in name_list if name_to_match.lower() in name.lower()]
    if len(match_list) > 1:
        # try requiring exact match
        match_list_temp = [
            name for name in match_list if name_to_match.lower() == name.lower()
        ]
        if len(match_list_temp) == 1:
            match_list = match_list_temp
        elif exclude_start_pattern is not None:
            # try excluding columns which start with exclude_start_pattern
            match_list_temp = [
                name
                for name in match_list
                if not name.startswith(exclude_start_pattern)
            ]
            if len(match_list_temp) == 1:
                match_list = match_list_temp
        if len(match_list) > 1:
            if brittle:
                raise ValueError(
                    f"More than one match for {name_to_match} in {name_list}."
                )
            else:
                return match_list
    elif len(match_list) == 0:
        if brittle:
            raise ValueError(f"No match for {name_to_match} in {name_list}.")
        else:
            return None
    return match_list[0]


def read_votable_catalog(
    votable_file,
    table_name="GLEAM",
    id_column="GLEAM",
    ra_column="RAJ2000",
    dec_column="DEJ2000",
    flux_columns="Fintwide",
    reference_frequency=200e6 * units.Hz,
    freq_array=None,
    spectral_index_column=None,
    source_select_kwds=None,
    return_table=False,
):
    """
    Create a SkyModel object from a votable catalog.

    Tested on: GLEAM EGC catalog, version 2

    Parameters
    ----------
    votable_file : str
        Path to votable catalog file.
    table_name : str
        Part of expected table name. Should match only one table name in votable_file.
    id_column : str
        Part of expected ID column. Should match only one column in the table.
    ra_column : str
        Part of expected RA column. Should match only one column in the table.
    dec_column : str
        Part of expected Dec column. Should match only one column in the table.
    flux_columns : str or list of str
        Part of expected Flux column(s). Each one should match only one column in the table.
    reference_frequency : astropy Quantity
        Reference frequency for flux values, assumed to be the same value for all rows.
    freq_array : astropy Quantity
        Frequencies corresponding to flux_columns (should be same length).
        Required for multiple flux columns.
    return_table : bool, optional
        Whether to return the astropy table instead of a list of Source objects.
    source_select_kwds : dict, optional
        Dictionary of keywords for source selection Valid options:

        * `lst_array`: For coarse RA horizon cuts, lsts used in the simulation [radians]
        * `latitude_deg`: Latitude of telescope in degrees. Used for declination coarse
           horizon cut.
        * `horizon_buffer`: Angle (float, in radians) of buffer for coarse horizon cut.
          Default is about 10 minutes of sky rotation. (See caveats in
          :func:`array_to_skymodel` docstring)
        * `min_flux`: Minimum stokes I flux to select [Jy]
        * `max_flux`: Maximum stokes I flux to select [Jy]

    Returns
    -------
    recarray or :class:`pyradiosky.SkyModel`
        if return_table, recarray of source parameters, otherwise :class:`pyradiosky.SkyModel` instance
    """
    parsed_vo = votable.parse(votable_file)

    tables = list(parsed_vo.iter_tables())
    table_ids = [table._ID for table in tables]

    if None not in table_ids:
        table_name_use = _get_matching_fields(table_name, table_ids)
        table_match = [table for table in tables if table._ID == table_name_use][0]
    else:
        warnings.warn(
            f"File {votable_file} contains tables with no name or ID, Support for "
            "such files is deprecated.",
            category=DeprecationWarning,
        )
        # Find correct table using the field names
        tables_match = []
        for table in tables:
            id_col_use = _get_matching_fields(
                id_column, table.to_table().colnames, brittle=False
            )
            if id_col_use is not None:
                tables_match.append(table)
        if len(tables_match) > 1:
            raise ValueError("More than one matching table.")
        else:
            table_match = tables_match[0]

    # Convert to astropy Table
    astropy_table = table_match.to_table()

    # get ID column
    id_col_use = _get_matching_fields(id_column, astropy_table.colnames)

    # get RA & Dec columns, if multiple matches, exclude VizieR calculated columns
    # which start with an underscore
    ra_col_use = _get_matching_fields(
        ra_column, astropy_table.colnames, exclude_start_pattern="_"
    )
    dec_col_use = _get_matching_fields(
        dec_column, astropy_table.colnames, exclude_start_pattern="_"
    )

    if isinstance(flux_columns, (str)):
        flux_columns = [flux_columns]
    flux_cols_use = []
    for col in flux_columns:
        flux_cols_use.append(_get_matching_fields(col, astropy_table.colnames))

    if len(flux_columns) > 1 and freq_array is None:
        raise ValueError("freq_array must be provided for multiple flux columns.")

    if reference_frequency is not None or len(flux_cols_use) == 1:
        if reference_frequency is not None:
            if not isinstance(reference_frequency, (Quantity,)):
                raise ValueError("reference_frequency must be an astropy Quantity.")
            reference_frequency = (
                np.array([reference_frequency.value] * len(astropy_table))
                * reference_frequency.unit
            )
        if spectral_index_column is not None:
            spectral_type = "spectral_index"
            spec_index_col_use = _get_matching_fields(
                spectral_index_column, astropy_table.colnames
            )
            spectral_index = astropy_table[spec_index_col_use].data.data
        else:
            spectral_type = "flat"
            spectral_index = None
    else:
        spectral_type = "subband"
        spectral_index = None

    stokes = np.zeros((4, len(flux_cols_use), len(astropy_table)), dtype=np.float)
    for index, col in enumerate(flux_cols_use):
        stokes[0, index, :] = astropy_table[col].data.data

    skymodel_obj = SkyModel(
        astropy_table[id_col_use].data.data.astype("str"),
        Longitude(astropy_table[ra_col_use].data.data, astropy_table[ra_col_use].unit),
        Latitude(astropy_table[dec_col_use].data.data, astropy_table[dec_col_use].unit),
        stokes,
        spectral_type,
        freq_array=freq_array,
        reference_frequency=reference_frequency,
        spectral_index=spectral_index,
    )

    if source_select_kwds is not None:
        sky_array = skymodel_to_array(skymodel_obj)
        sky_array = source_cuts(sky_array, **source_select_kwds)
        if return_table:
            return sky_array

        skymodel_obj = array_to_skymodel(sky_array)

    if return_table:
        return skymodel_to_array(skymodel_obj)

    return skymodel_obj


def read_gleam_catalog(
    votable_file, spectral_type="flat", source_select_kwds=None, return_table=False
):
    """
    Create a SkyModel object from the GLEAM votable catalog.

    Tested on: GLEAM EGC catalog, version 2

    Parameters
    ----------
    votable_file : str
        Path to GLEAM votable catalog file.
    spectral_type : str
        One of 'flat', 'subband' or 'spectral_index'. If set to 'flat', the
        wide band integrated flux will be used, if set to 'spectral_index' the
        fitted flux at 200 MHz will be used for the flux column.
    return_table : bool, optional
        Whether to return the astropy table instead of a SkyModel object.
    source_select_kwds : dict, optional
        Dictionary of keywords for source selection Valid options:

        * `lst_array`: For coarse RA horizon cuts, lsts used in the simulation [radians]
        * `latitude_deg`: Latitude of telescope in degrees. Used for declination coarse
           horizon cut.
        * `horizon_buffer`: Angle (float, in radians) of buffer for coarse horizon cut.
          Default is about 10 minutes of sky rotation. (See caveats in
          :func:`array_to_skymodel` docstring)
        * `min_flux`: Minimum stokes I flux to select [Jy]
        * `max_flux`: Maximum stokes I flux to select [Jy]

    Returns
    -------
    recarray or :class:`pyradiosky.SkyModel`
        if return_table, recarray of source parameters, otherwise :class:`pyradiosky.SkyModel` instance
    """
    spec_type_list = ["flat", "spectral_index", "subband"]
    if spectral_type not in spec_type_list:
        raise ValueError(
            f"spectral_type {spectral_type} is not an allowed type. "
            f"Allowed types are: {spec_type_list}"
        )

    if spectral_type == "flat":
        flux_columns = "Fintwide"
        reference_frequency = 200e6 * units.Hz
        freq_array = None
        spectral_index_column = None
    elif spectral_type == "spectral_index":
        flux_columns = "Fintfit200"
        reference_frequency = 200e6 * units.Hz
        spectral_index_column = "alpha"
        freq_array = None
    else:
        # fmt: off
        flux_columns = ["Fint076", "Fint084", "Fint092", "Fint099", "Fint107",
                        "Fint115", "Fint122", "Fint130", "Fint143", "Fint151",
                        "Fint158", "Fint166", "Fint174", "Fint181", "Fint189",
                        "Fint197", "Fint204", "Fint212", "Fint220", "Fint227"]
        freq_array = [76, 84, 92, 99, 107, 115, 122, 130, 143, 151, 158, 166,
                      174, 181, 189, 197, 204, 212, 220, 227]
        freq_array = np.array(freq_array) * 1e6 * units.Hz
        reference_frequency = None
        spectral_index_column = None
        # fmt: on

    gleam_skymodel = read_votable_catalog(
        votable_file,
        table_name="GLEAM",
        id_column="GLEAM",
        ra_column="RAJ2000",
        dec_column="DEJ2000",
        flux_columns=flux_columns,
        freq_array=freq_array,
        reference_frequency=reference_frequency,
        spectral_index_column=spectral_index_column,
        source_select_kwds=source_select_kwds,
        return_table=return_table,
    )

    return gleam_skymodel


def read_text_catalog(catalog_csv, source_select_kwds=None, return_table=False):
    """
    Read in a text file of sources.

    Parameters
    ----------
    catalog_csv: str
        Path to tab separated value file with the following required columns:
        *  `Source_ID`: source name as a string of maximum 10 characters
        *  `ra_j2000`: right ascension at J2000 epoch, in decimal degrees
        *  `dec_j2000`: declination at J2000 epoch, in decimal degrees
        *  `Flux [Jy]`: Stokes I flux density in Janskys

        If flux is specified at multiple frequencies (must be the same set for all
        components), the frequencies must be included in each column name,
        e.g. `Flux at 150 MHz [Jy]`. Recognized units are ('Hz', 'kHz', 'MHz' or 'GHz'):

        If flux is only specified at one reference frequency (can be different per
        component), a frequency column should be added (note: assumed to be in Hz):
        *  `Frequency`: reference frequency [Hz]

        Optionally a spectral index can be specified per component with:
        *  `Spectral_Index`: spectral index

    source_select_kwds : dict, optional
        Dictionary of keywords for source selection. Valid options:

        * `lst_array`: For coarse RA horizon cuts, lsts used in the simulation [radians]
        * `latitude_deg`: Latitude of telescope in degrees. Used for declination coarse
        *  horizon cut.
        * `horizon_buffer`: Angle (float, in radians) of buffer for coarse horizon cut.
          Default is about 10 minutes of sky rotation. (See caveats in
          :func:`array_to_skymodel` docstring)
        * `min_flux`: Minimum stokes I flux to select [Jy]
        * `max_flux`: Maximum stokes I flux to select [Jy]

    Returns
    -------
    sky_model : :class:`SkyModel`
        A sky model created from the text catalog.
    """
    with open(catalog_csv, "r") as cfile:
        header = cfile.readline()
    header = [
        h.strip() for h in header.split() if not h[0] == "["
    ]  # Ignore units in header

    flux_fields = [colname for colname in header if colname.lower().startswith("flux")]
    flux_fields_lower = [colname.lower() for colname in flux_fields]

    header_lower = [colname.lower() for colname in header]

    expected_cols = ["source_id", "ra_j2000", "dec_j2000"]
    if "frequency" in header_lower:
        if len(flux_fields) != 1:
            raise ValueError(
                "If frequency column is present, only one flux column allowed."
            )
        freq_array = None
        expected_cols.extend([flux_fields_lower[0], "frequency"])
        if "spectral_index" in header_lower:
            spectral_type = "spectral_index"
            expected_cols.append("spectral_index")
            freq_array = None
        else:
            spectral_type = "flat"
        n_freqs = 1
    else:
        frequencies = []
        for fluxname in flux_fields:
            if "Hz" in fluxname:
                cst_obj = CSTBeam()
                freq = cst_obj.name2freq(fluxname)
                frequencies.append(freq)
            else:
                if len(flux_fields) > 1:
                    raise ValueError(
                        "Multiple flux fields, but they do not all contain a frequency."
                    )
        if len(frequencies) > 0:
            n_freqs = len(frequencies)
            if "subband" in flux_fields[0]:
                spectral_type = "subband"
            else:
                if len(frequencies) > 1:
                    spectral_type = "full"
                else:
                    spectral_type = "flat"
            # This has a freq_array
            expected_cols.extend(flux_fields_lower)
            freq_array = np.array(frequencies) * units.Hz
        else:
            # This is a flat spectrum (no freq info)
            n_freqs = 1
            spectral_type = "flat"
            freq_array = None
            expected_cols.append("flux")

    if expected_cols != header_lower:
        raise ValueError(
            "Header does not match expectations. Expected columns"
            f"are: {expected_cols}, header columns were: {header_lower}"
        )

    catalog_table = np.genfromtxt(
        catalog_csv, autostrip=True, skip_header=1, dtype=None, encoding="utf-8"
    )

    catalog_table = np.atleast_1d(catalog_table)

    col_names = catalog_table.dtype.names

    names = catalog_table[col_names[0]].astype("str")
    ras = Longitude(catalog_table[col_names[1]], units.deg)
    decs = Latitude(catalog_table[col_names[2]], units.deg)

    stokes = np.zeros((4, n_freqs, len(catalog_table)), dtype=np.float)
    for ind in np.arange(n_freqs):
        stokes[0, ind, :] = catalog_table[col_names[ind + 3]]

    if "frequency" in header_lower and freq_array is None:
        freq_ind = np.where(np.array(header_lower) == "frequency")[0][0]
        reference_frequency = catalog_table[col_names[freq_ind]] * units.Hz
        if "spectral_index" in header_lower:
            si_ind = np.where(np.array(header_lower) == "spectral_index")[0][0]
            spectral_index = catalog_table[col_names[si_ind]]
        else:
            spectral_index = None
    else:
        reference_frequency = None
        spectral_index = None

    skymodel_obj = SkyModel(
        names,
        ras,
        decs,
        stokes,
        spectral_type,
        freq_array=freq_array,
        reference_frequency=reference_frequency,
        spectral_index=spectral_index,
    )

    if source_select_kwds is not None:
        sky_array = skymodel_to_array(skymodel_obj)
        sky_array = source_cuts(sky_array, **source_select_kwds)
        if return_table:
            return sky_array

        skymodel_obj = array_to_skymodel(sky_array)

    if return_table:
        return skymodel_to_array(skymodel_obj)

    return skymodel_obj


def read_idl_catalog(filename_sav, expand_extended=True):
    """
    Read in an FHD-readable IDL .sav file catalog.

    Parameters
    ----------
    filename_sav: str
        Path to IDL .sav file.

    expand_extended: bool
        If True, return extended source components.
        Default: True

    Returns
    -------
    :class:`pyradiosky.SkyModel`
    """
    catalog = scipy.io.readsav(filename_sav)["catalog"]
    ids = catalog["id"].astype(str)
    ra = catalog["ra"]
    dec = catalog["dec"]
    source_freqs = catalog["freq"]
    spectral_index = catalog["alpha"]
    Nsrcs = len(catalog)
    extended_model_group = np.full(Nsrcs, -1, dtype=int)
    if "BEAM" in catalog.dtype.names:
        use_beam_amps = True
        beam_amp = np.zeros((4, Nsrcs))
    else:
        use_beam_amps = False
        beam_amp = None
    stokes = np.zeros((4, Nsrcs))
    for src in range(Nsrcs):
        stokes[0, src] = catalog["flux"][src]["I"][0]
        stokes[1, src] = catalog["flux"][src]["Q"][0]
        stokes[2, src] = catalog["flux"][src]["U"][0]
        stokes[3, src] = catalog["flux"][src]["V"][0]
        if use_beam_amps:
            beam_amp[0, src] = catalog["beam"][src]["XX"][0]
            beam_amp[1, src] = catalog["beam"][src]["YY"][0]
            beam_amp[2, src] = catalog["beam"][src]["XY"][0]
            beam_amp[3, src] = catalog["beam"][src]["YX"][0]

    if expand_extended:
        ext_inds = np.where(
            [catalog["extend"][ind] is not None for ind in range(Nsrcs)]
        )[0]
        source_group_id = 1
        if len(ext_inds) > 0:  # Add components and preserve ordering
            source_inds = np.array(range(Nsrcs))
            for ext in ext_inds:
                use_index = np.where(source_inds == ext)[0][0]
                source_id = ids[use_index]
                # Remove top-level source information
                ids = np.delete(ids, ext)
                ra = np.delete(ra, ext)
                dec = np.delete(dec, ext)
                stokes = np.delete(stokes, ext, axis=1)
                source_freqs = np.delete(source_freqs, ext)
                spectral_index = np.delete(spectral_index, ext)
                source_inds = np.delete(source_inds, ext)
                # Add component information
                src = catalog[ext]["extend"]
                Ncomps = len(src)
                ids = np.insert(ids, use_index, np.full(Ncomps, source_id))
                extended_model_group = np.insert(
                    extended_model_group, use_index, np.full(Ncomps, source_group_id)
                )
                source_group_id += 1
                ra = np.insert(ra, use_index, src["ra"])
                dec = np.insert(dec, use_index, src["dec"])
                stokes_ext = np.zeros((4, Ncomps))
                if use_beam_amps:
                    beam_amp_ext = np.zeros((4, Ncomps))
                for comp in range(Ncomps):
                    stokes_ext[0, comp] = src["flux"][comp]["I"][0]
                    stokes_ext[1, comp] = src["flux"][comp]["Q"][0]
                    stokes_ext[2, comp] = src["flux"][comp]["U"][0]
                    stokes_ext[3, comp] = src["flux"][comp]["V"][0]
                    if use_beam_amps:
                        beam_amp_ext[0, comp] = src["beam"][comp]["XX"][0]
                        beam_amp_ext[1, comp] = src["beam"][comp]["YY"][0]
                        beam_amp_ext[2, comp] = src["beam"][comp]["XY"][0]
                        beam_amp_ext[3, comp] = src["beam"][comp]["YX"][0]
                stokes = np.concatenate(
                    (  # np.insert doesn't work with arrays
                        np.concatenate((stokes[:, :use_index], stokes_ext), axis=1),
                        stokes[:, use_index:],
                    ),
                    axis=1,
                )
                if use_beam_amps:
                    beam_amp = np.concatenate(
                        (
                            np.concatenate(
                                (beam_amp[:, :use_index], beam_amp_ext), axis=1
                            ),
                            beam_amp[:, use_index:],
                        ),
                        axis=1,
                    )
                source_freqs = np.insert(source_freqs, use_index, src["freq"])
                spectral_index = np.insert(spectral_index, use_index, src["alpha"])
                source_inds = np.insert(source_inds, use_index, np.full(Ncomps, ext))

    ra = Longitude(ra, units.deg)
    dec = Latitude(dec, units.deg)
    stokes = stokes[:, np.newaxis, :]  # Add frequency axis
    sourcelist = SkyModel(
        ids,
        ra,
        dec,
        stokes,
        spectral_type="spectral_index",
        reference_frequency=Quantity(source_freqs, "hertz"),
        spectral_index=spectral_index,
        beam_amp=beam_amp,
        extended_model_group=extended_model_group,
    )
    return sourcelist


def write_catalog_to_file(filename, skymodel):
    """
    Write out a catalog to a text file.

    Readable with :meth:`read_text_catalog()`.

    Parameters
    ----------
    filename : str
        Path to output file (string)
    skymodel : :class:`SkyModel`
        The sky model to write to file.
    """
    header = "SOURCE_ID\tRA_J2000 [deg]\tDec_J2000 [deg]"
    format_str = "{}\t{:0.8f}\t{:0.8f}"
    if skymodel.reference_frequency is not None:
        header += "\tFlux [Jy]\tFrequency [Hz]"
        format_str += "\t{:0.8f}\t{:0.8f}"
        if skymodel.spectral_index is not None:
            header += "\tSpectral_Index"
            format_str += "\t{:0.8f}"
    elif skymodel.freq_array is not None:
        for freq in skymodel.freq_array:
            freq_hz_val = freq.to(units.Hz).value
            if freq_hz_val > 1e9:
                freq_str = "{:g}_GHz".format(freq_hz_val * 1e-9)
            elif freq_hz_val > 1e6:
                freq_str = "{:g}_MHz".format(freq_hz_val * 1e-6)
            elif freq_hz_val > 1e3:
                freq_str = "{:g}_kHz".format(freq_hz_val * 1e-3)
            else:
                freq_str = "{:g}_Hz".format(freq_hz_val)

            if skymodel.spectral_type == "subband":
                header += f"\tFlux_subband_{freq_str} [Jy]"
            else:
                header += f"\tFlux_{freq_str} [Jy]"
            format_str += "\t{:0.8f}"
    else:
        # flat spectral response, no freq info
        header += "\tFlux [Jy]"
        format_str += "\t{:0.8f}"

    header += "\n"
    format_str += "\n"

    with open(filename, "w+") as fo:
        fo.write(header)
        arr = skymodel_to_array(skymodel)
        for src in arr:
            if skymodel.reference_frequency is not None:
                if skymodel.spectral_index is not None:
                    srcid, ra, dec, flux, rfreq, freq, spec_index = src
                    flux_i = flux[0]
                    fo.write(
                        format_str.format(srcid, ra, dec, flux_i, freq, spec_index)
                    )
                else:
                    srcid, ra, dec, flux, rfreq, freq = src
                    flux_i = flux[0]
                    fo.write(format_str.format(srcid, ra, dec, flux_i, rfreq))
            elif skymodel.freq_array is not None:
                srcid, ra, dec, flux, freq = src
                flux_i = flux[..., 0]
                fo.write(format_str.format(srcid, ra, dec, *flux_i))
            else:
                # flat spectral response, no freq info
                srcid, ra, dec, flux = src
                flux_i = flux[0]
                fo.write(format_str.format(srcid, ra, dec, flux_i))



"""Define namespace."""
from setuptools_scm import get_version
from pathlib import Path
from pkg_resources import get_distribution, DistributionNotFound

from .branch_scheme import branch_scheme


try:  # pragma: nocover
    # get accurate version for developer installs
    version_str = get_version(Path(__file__).parent.parent, local_scheme=branch_scheme)

    __version__ = version_str

except (LookupError, ImportError):
    try:
        # Set the version automatically from the package details.
        __version__ = get_distribution(__name__).version
    except DistributionNotFound:  # pragma: nocover
        # package is not installed
        pass

from .skymodel import (
    SkyModel,
    read_healpix_hdf5,
    healpix_to_sky,
    skymodel_to_array,
    array_to_skymodel,
    source_cuts,
    read_votable_catalog,
    read_text_catalog,
    read_idl_catalog,
    write_catalog_to_file,
    write_healpix_hdf5,
)



# -*- mode: python; coding: utf-8 -*
# Copyright (c) 2019 Radio Astronomy Software Group
# Licensed under the 3-clause BSD License
"""Utility methods."""

import numpy as np
from astropy.constants import c
from astropy import _erfa as erfa
from astropy.coordinates.builtin_frames.utils import get_jd12
from astropy.time import Time
from astropy.coordinates import Angle


# The frame radio astronomers call the apparent or current epoch is the
# "true equator & equinox" frame, notated E_upsilon in the USNO circular
# astropy doesn't have this frame but it's pretty easy to adapt the CIRS frame
# by modifying the ra to reflect the difference between
# GAST (Grenwich Apparent Sidereal Time) and the earth rotation angle (theta)
def _tee_to_cirs_ra(tee_ra, time):
    """
    Convert from the true equator & equinox frame to the CIRS frame.

    The frame radio astronomers call the apparent or current epoch is the
    "true equator & equinox" frame, notated E_upsilon in the USNO circular
    astropy doesn't have this frame but it's pretty easy to adapt the CIRS frame
    by modifying the ra to reflect the difference between
    GAST (Grenwich Apparent Sidereal Time) and the earth rotation angle (theta)

    Parameters
    ----------
    tee_ra : :class:`astropy.Angle`
        Current epoch RA (RA in the true equator and equinox frame).
    time : :class:`astropy.Time`
        Time object for the epoch of the `tee_ra`.
    """
    era = erfa.era00(*get_jd12(time, "ut1"))
    theta_earth = Angle(era, unit="rad")

    assert isinstance(time, Time)
    assert isinstance(tee_ra, Angle)
    gast = time.sidereal_time("apparent", longitude=0)
    cirs_ra = tee_ra - (gast - theta_earth)
    return cirs_ra


def _cirs_to_tee_ra(cirs_ra, time):
    """
    Convert from CIRS frame to the true equator & equinox frame.

    The frame radio astronomers call the apparent or current epoch is the
    "true equator & equinox" frame, notated E_upsilon in the USNO circular
    astropy doesn't have this frame but it's pretty easy to adapt the CIRS frame
    by modifying the ra to reflect the difference between
    GAST (Grenwich Apparent Sidereal Time) and the earth rotation angle (theta)

    Parameters
    ----------
    cirs_ra : :class:`astropy.Angle`
        CIRS RA.
    time : :class:`astropy.Time`
        Time object for time to convert to the "true equator & equinox" frame.
    """
    era = erfa.era00(*get_jd12(time, "ut1"))
    theta_earth = Angle(era, unit="rad")

    assert isinstance(time, Time)
    assert isinstance(cirs_ra, Angle)
    gast = time.sidereal_time("apparent", longitude=0)
    tee_ra = cirs_ra + (gast - theta_earth)
    return tee_ra


def stokes_to_coherency(stokes_vector):
    """
    Convert Stokes vector to coherency matrix.

    Parameters
    ----------
    stokes_vector : array_like of float
        Vector(s) of stokes parameters in order [I, Q, U, V], shape(4,) or (4, Nfreqs, Ncomponents)

    Returns
    -------
    coherency matrix : array of float
        Array of coherencies, shape (2, 2) or (2, 2, Nfreqs, Ncomponents)
    """
    stokes_arr = np.atleast_1d(np.asarray(stokes_vector))
    initial_shape = stokes_arr.shape
    if initial_shape[0] != 4:
        raise ValueError("First dimension of stokes_vector must be length 4.")

    if stokes_arr.size == 4 and len(initial_shape) == 1:
        stokes_arr = stokes_arr[:, np.newaxis, np.newaxis]

    coherency = 0.5 * np.array(
        [
            [
                stokes_arr[0, :, :] + stokes_arr[1, :, :],
                stokes_arr[2, :, :] - 1j * stokes_arr[3, :, :],
            ],
            [
                stokes_arr[2, :, :] + 1j * stokes_arr[3, :, :],
                stokes_arr[0, :, :] - stokes_arr[1, :, :],
            ],
        ]
    )

    if stokes_arr.size == 4 and len(initial_shape) == 1:
        coherency = np.squeeze(coherency)
    return coherency


def coherency_to_stokes(coherency_matrix):
    """
    Convert coherency matrix to vector of 4 Stokes parameter in order [I, Q, U, V].

    Parameters
    ----------
    coherency matrix : array_like of float
        Array of coherencies, shape (2, 2) or (2, 2, Ncomponents)

    Returns
    -------
    stokes_vector : array of float
        Vector(s) of stokes parameters, shape(4,) or (4, Ncomponents)
    """
    coherency_arr = np.asarray(coherency_matrix)
    initial_shape = coherency_arr.shape
    if len(initial_shape) < 2 or initial_shape[0] != 2 or initial_shape[1] != 2:
        raise ValueError("First two dimensions of coherency_matrix must be length 2.")

    if coherency_arr.size == 4 and len(initial_shape) == 2:
        coherency_arr = coherency_arr[:, :, np.newaxis]

    stokes = np.array(
        [
            coherency_arr[0, 0, :] + coherency_arr[1, 1, :],
            coherency_arr[0, 0, :] - coherency_arr[1, 1, :],
            coherency_arr[0, 1, :] + coherency_arr[1, 0, :],
            -(coherency_arr[0, 1, :] - coherency_arr[1, 0, :]).imag,
        ]
    ).real
    if coherency_arr.size == 4 and len(initial_shape) == 2:
        stokes = np.squeeze(stokes)

    return stokes


def jy_to_ksr(freqs):
    """
    Calculate multiplicative factors to convert [Jy] to [K sr].

    Parameters
    ----------
    freqs : array_like of float
        Frequencies in Hz.
    """
    c_cmps = c.to("cm/s").value  # cm/s
    k_boltz = 1.380658e-16  # erg/K
    lambdas = c_cmps / freqs  # cm
    return 1e-23 * lambdas ** 2 / (2 * k_boltz)



# -*- mode: python; coding: utf-8 -*
# Copyright (c) 2019 Radio Astronomy Software Group
# Licensed under the 3-clause BSD License

import numpy as np
import pytest
from astropy.coordinates import Angle
from astropy.time import Time

import pyradiosky.utils as skyutils


def test_tee_ra_loop():
    time = Time(2457458.1739, scale="utc", format="jd")
    tee_ra = Angle(np.pi / 4.0, unit="rad")  # rad
    cirs_ra = skyutils._tee_to_cirs_ra(tee_ra, time)
    new_tee_ra = skyutils._cirs_to_tee_ra(cirs_ra, time)
    assert new_tee_ra == tee_ra


def test_stokes_tofrom_coherency():

    stokesI = 4.5
    stokesQ = -0.3
    stokesU = 1.2
    stokesV = -0.15
    stokes = np.array([stokesI, stokesQ, stokesU, stokesV])

    expected_coherency = 0.5 * np.array([[4.2, 1.2 + 0.15j], [1.2 - 0.15j, 4.8]])

    coherency = skyutils.stokes_to_coherency(stokes)

    assert np.allclose(expected_coherency, coherency.squeeze())

    back_to_stokes = skyutils.coherency_to_stokes(coherency)

    assert np.allclose(stokes, back_to_stokes)

    # again, with multiple sources and a frequency axis.
    stokes = np.array(
        [[stokesI, stokesQ, stokesU, stokesV], [stokesI, stokesQ, stokesU, stokesV]]
    ).T

    stokes = stokes[:, np.newaxis, :]

    coherency = skyutils.stokes_to_coherency(stokes)
    back_to_stokes = skyutils.coherency_to_stokes(coherency)

    assert np.allclose(stokes, back_to_stokes)

    with pytest.raises(ValueError) as cm:
        skyutils.stokes_to_coherency(stokes[0:2, :])
    assert str(cm.value).startswith(
        "First dimension of stokes_vector must be length 4."
    )

    with pytest.raises(ValueError) as cm:
        skyutils.coherency_to_stokes(expected_coherency[0, :])
    assert str(cm.value).startswith(
        "First two dimensions of coherency_matrix must be length 2."
    )
